@@ -24,6 +24,7 @@
 #include <poll.h>
 
 #include "commands/dbcommands.h"
+#include "distributed/connection_management.h"
 #include "distributed/multi_client_executor.h"
 #include "distributed/multi_physical_planner.h"
 #include "distributed/multi_server_executor.h"
@@ -81,7 +82,7 @@ MultiRealTimeExecute(Job *job)
 	const char *workerHashName = "Worker node hash";
 	WaitInfo *waitInfo = MultiClientCreateWaitInfo(list_length(taskList));
 
-	workerNodeList = WorkerNodeList();
+	workerNodeList = ActiveWorkerNodeList();
 	workerHash = WorkerHash(workerHashName, workerNodeList);
 
 	/* initialize task execution structures for remote execution */
@@ -224,8 +225,7 @@ MultiRealTimeExecute(Job *job)
 	 */
 	if (taskFailed)
 	{
-		ereport(ERROR, (errmsg("failed to execute job " UINT64_FORMAT, job->jobId),
-						errdetail("Failure due to failed task %u", failedTaskId)));
+		ereport(ERROR, (errmsg("failed to execute task %u", failedTaskId)));
 	}
 	else if (QueryCancelPending)
 	{
@@ -332,11 +332,11 @@ ManageTaskExecution(Task *task, TaskExecution *taskExecution,
 			{
 				if (TimestampDifferenceExceeds(taskExecution->connectStartTime,
 											   GetCurrentTimestamp(),
-											   REMOTE_NODE_CONNECT_TIMEOUT))
+											   NodeConnectionTimeout))
 				{
 					ereport(WARNING, (errmsg("could not establish asynchronous "
 											 "connection after %u ms",
-											 REMOTE_NODE_CONNECT_TIMEOUT)));
+											 NodeConnectionTimeout)));
 
 					taskStatusArray[currentIndex] = EXEC_TASK_FAILED;
 				}
