@@ -18,7 +18,7 @@
 
 #include "access/htup_details.h"
 #include "catalog/pg_type.h"
-#include "distributed/connection_cache.h"
+#include "distributed/connection_management.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/multi_server_executor.h"
 #include "distributed/worker_protocol.h"
@@ -83,6 +83,8 @@ master_run_on_worker(PG_FUNCTION_ARGS)
 	int commandIndex = 0;
 	int commandCount = 0;
 
+	CheckCitusVersion(ERROR);
+
 	/* check to see if caller supports us returning a tuplestore */
 	if (!rsinfo || !(rsinfo->allowedModes & SFRM_Materialize))
 	{
@@ -221,7 +223,6 @@ ExecuteCommandsInParallelAndStoreResults(StringInfo *nodeNameArray, int *nodePor
 										 int commmandCount)
 {
 	int commandIndex = 0;
-	char *nodeUser = CurrentUserName();
 	PGconn **connectionArray = palloc0(commmandCount * sizeof(PGconn *));
 	int finishedCount = 0;
 
@@ -230,20 +231,25 @@ ExecuteCommandsInParallelAndStoreResults(StringInfo *nodeNameArray, int *nodePor
 	{
 		char *nodeName = nodeNameArray[commandIndex]->data;
 		int nodePort = nodePortArray[commandIndex];
-		PGconn *connection = ConnectToNode(nodeName, nodePort, nodeUser);
+		int connectionFlags = FORCE_NEW_CONNECTION;
+		MultiConnection *multiConnection =
+			GetNodeConnection(connectionFlags, nodeName, nodePort);
+		PGconn *connection = multiConnection->pgConn;
 		StringInfo queryResultString = resultStringArray[commandIndex];
 
 		statusArray[commandIndex] = true;
 
-		connectionArray[commandIndex] = connection;
-
-		if (connection == NULL)
+		if (PQstatus(multiConnection->pgConn) != CONNECTION_OK)
 		{
 			appendStringInfo(queryResultString, "failed to connect to %s:%d", nodeName,
 							 (int) nodePort);
 			statusArray[commandIndex] = false;
 			finishedCount++;
 		}
+		else
+		{
+			connectionArray[commandIndex] = connection;
+		}
 	}
 
 	/* send queries at once */
@@ -269,7 +275,7 @@ ExecuteCommandsInParallelAndStoreResults(StringInfo *nodeNameArray, int *nodePor
 		{
 			StoreErrorMessage(connection, queryResultString);
 			statusArray[commandIndex] = false;
-			PQfinish(connection);
+			CloseConnectionByPGconn(connection);
 			connectionArray[commandIndex] = NULL;
 			finishedCount++;
 		}
@@ -298,7 +304,7 @@ ExecuteCommandsInParallelAndStoreResults(StringInfo *nodeNameArray, int *nodePor
 				finishedCount++;
 				statusArray[commandIndex] = success;
 				connectionArray[commandIndex] = NULL;
-				PQfinish(connection);
+				CloseConnectionByPGconn(connection);
 			}
 		}
 
@@ -431,17 +437,19 @@ static void
 StoreErrorMessage(PGconn *connection, StringInfo queryResultString)
 {
 	char *errorMessage = PQerrorMessage(connection);
-	char *firstNewlineIndex = strchr(errorMessage, '\n');
-
-	/* trim the error message at the line break */
-	if (firstNewlineIndex != NULL)
+	if (errorMessage != NULL)
 	{
-		*firstNewlineIndex = '\0';
+		char *firstNewlineIndex = strchr(errorMessage, '\n');
+
+		/* trim the error message at the line break */
+		if (firstNewlineIndex != NULL)
+		{
+			*firstNewlineIndex = '\0';
+		}
 	}
-
-	/* put a default error message if no error message is reported */
-	if (errorMessage == NULL)
+	else
 	{
+		/* put a default error message if no error message is reported */
 		errorMessage = "An error occurred while running the query";
 	}
 
@@ -490,37 +498,27 @@ static bool
 ExecuteRemoteQueryOrCommand(char *nodeName, uint32 nodePort, char *queryString,
 							StringInfo queryResultString)
 {
-	char *nodeUser = CurrentUserName();
-	PGconn *nodeConnection = ConnectToNode(nodeName, nodePort, nodeUser);
+	int connectionFlags = FORCE_NEW_CONNECTION;
+	MultiConnection *multiConnection =
+		GetNodeConnection(connectionFlags, nodeName, nodePort);
+	PGconn *nodeConnection = multiConnection->pgConn;
 	bool success = false;
+	PGresult *queryResult = NULL;
 
-	if (nodeConnection == NULL)
+	if (PQstatus(multiConnection->pgConn) != CONNECTION_OK)
 	{
 		appendStringInfo(queryResultString, "failed to connect to %s:%d", nodeName,
 						 (int) nodePort);
 		return false;
 	}
 
-	PG_TRY();
-	{
-		PGresult *queryResult = PQexec(nodeConnection, queryString);
-		success = EvaluateQueryResult(nodeConnection, queryResult, queryResultString);
+	queryResult = PQexec(nodeConnection, queryString);
+	success = EvaluateQueryResult(nodeConnection, queryResult, queryResultString);
 
-		PQclear(queryResult);
+	PQclear(queryResult);
 
-		/* close the connection */
-		PQfinish(nodeConnection);
-		nodeConnection = NULL;
-	}
-	PG_CATCH();
-	{
-		StoreErrorMessage(nodeConnection, queryResultString);
-
-		/* close the connection */
-		PQfinish(nodeConnection);
-		nodeConnection = NULL;
-	}
-	PG_END_TRY();
+	/* close the connection */
+	CloseConnection(multiConnection);
 
 	return success;
 }
