@@ -16,6 +16,7 @@
 #include "postgres.h"
 #include "c.h"
 #include "fmgr.h"
+#include "libpq-fe.h"
 #include "miscadmin.h"
 #include "port.h"
 
@@ -24,16 +25,22 @@
 #include "access/xact.h"
 #include "catalog/namespace.h"
 #include "commands/dbcommands.h"
+#include "distributed/connection_management.h"
 #include "distributed/master_protocol.h"
+#include "distributed/metadata_sync.h"
 #include "distributed/multi_client_executor.h"
 #include "distributed/multi_join_order.h"
 #include "distributed/multi_logical_planner.h"
 #include "distributed/multi_physical_planner.h"
 #include "distributed/multi_server_executor.h"
+#include "distributed/multi_utility.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
+#include "distributed/placement_connection.h"
 #include "distributed/relay_utility.h"
+#include "distributed/remote_commands.h"
 #include "distributed/worker_protocol.h"
+#include "distributed/worker_transaction.h"
 #include "lib/stringinfo.h"
 #include "nodes/nodes.h"
 #include "nodes/parsenodes.h"
@@ -44,6 +51,7 @@
 #include "optimizer/predtest.h"
 #include "optimizer/restrictinfo.h"
 #include "storage/lock.h"
+#include "storage/lmgr.h"
 #include "tcop/tcopprot.h"
 #include "utils/array.h"
 #include "utils/builtins.h"
@@ -98,14 +106,12 @@ master_apply_delete_command(PG_FUNCTION_ARGS)
 	Node *queryTreeNode = NULL;
 	DeleteStmt *deleteStatement = NULL;
 	int droppedShardCount = 0;
-	LOCKTAG lockTag;
-	bool sessionLock = false;
-	bool dontWait = false;
+	LOCKMODE lockMode = 0;
 	char partitionMethod = 0;
 	bool failOK = false;
-	bool isTopLevel = true;
 
-	PreventTransactionChain(isTopLevel, "master_apply_delete_command");
+	EnsureCoordinator();
+	CheckCitusVersion(ERROR);
 
 	queryTreeNode = ParseTreeNode(queryString);
 	if (!IsA(queryTreeNode, DeleteStmt))
@@ -118,7 +124,15 @@ master_apply_delete_command(PG_FUNCTION_ARGS)
 
 	schemaName = deleteStatement->relation->schemaname;
 	relationName = deleteStatement->relation->relname;
-	relationId = RangeVarGetRelid(deleteStatement->relation, NoLock, failOK);
+
+	/*
+	 * We take an exclusive lock while dropping shards to prevent concurrent
+	 * writes. We don't want to block SELECTs, which means queries might fail
+	 * if they access a shard that has just been dropped.
+	 */
+	lockMode = ExclusiveLock;
+
+	relationId = RangeVarGetRelid(deleteStatement->relation, lockMode, failOK);
 
 	/* schema-prefix if it is not specified already */
 	if (schemaName == NULL)
@@ -139,21 +153,27 @@ master_apply_delete_command(PG_FUNCTION_ARGS)
 	deleteCriteria = eval_const_expressions(NULL, whereClause);
 
 	partitionMethod = PartitionMethod(relationId);
-	if ((partitionMethod == DISTRIBUTE_BY_HASH) && (deleteCriteria != NULL))
+	if (partitionMethod == DISTRIBUTE_BY_HASH)
+	{
+		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						errmsg("cannot delete from hash distributed table with this "
+							   "command"),
+						errdetail("Delete statements on hash-partitioned tables "
+								  "are not supported with master_apply_delete_command."),
+						errhint("Use master_modify_multiple_shards command instead.")));
+	}
+	else if (partitionMethod == DISTRIBUTE_BY_NONE)
 	{
 		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 						errmsg("cannot delete from distributed table"),
-						errdetail("Delete statements on hash-partitioned tables "
-								  "with where clause is not supported")));
+						errdetail("Delete statements on reference tables "
+								  "are not supported.")));
 	}
 
+
 	CheckDeleteCriteria(deleteCriteria);
 	CheckPartitionColumn(relationId, deleteCriteria);
 
-	/* acquire lock */
-	SET_LOCKTAG_ADVISORY(lockTag, MyDatabaseId, relationId, 0, 0);
-	LockAcquire(&lockTag, ExclusiveLock, sessionLock, dontWait);
-
 	shardIntervalList = LoadShardIntervalList(relationId);
 
 	/* drop all shards if where clause is not present */
@@ -188,45 +208,24 @@ master_drop_all_shards(PG_FUNCTION_ARGS)
 	text *schemaNameText = PG_GETARG_TEXT_P(1);
 	text *relationNameText = PG_GETARG_TEXT_P(2);
 
-	char *schemaName = NULL;
-	char *relationName = NULL;
-	bool isTopLevel = true;
 	List *shardIntervalList = NIL;
 	int droppedShardCount = 0;
 
-	PreventTransactionChain(isTopLevel, "DROP distributed table");
+	char *schemaName = text_to_cstring(schemaNameText);
+	char *relationName = text_to_cstring(relationNameText);
 
-	relationName = get_rel_name(relationId);
+	EnsureCoordinator();
+	CheckCitusVersion(ERROR);
 
-	if (relationName != NULL)
-	{
-		/* ensure proper values are used if the table exists */
-		Oid schemaId = get_rel_namespace(relationId);
-		schemaName = get_namespace_name(schemaId);
+	CheckTableSchemaNameForDrop(relationId, &schemaName, &relationName);
 
-		/*
-		 * Only allow the owner to drop all shards, this is more akin to DDL
-		 * than DELETE.
-		 */
-		EnsureTableOwner(relationId);
-	}
-	else
-	{
-		/* table has been dropped, rely on user-supplied values */
-		schemaName = text_to_cstring(schemaNameText);
-		relationName = text_to_cstring(relationNameText);
-
-		/*
-		 * Verify that this only is run as superuser - that's how it's used in
-		 * our drop event trigger, and we can't verify permissions for an
-		 * already dropped relation.
-		 */
-		if (!superuser())
-		{
-			ereport(ERROR, (errmsg("cannot drop all shards of a dropped table as "
-								   "non-superuser")));
-		}
-	}
+	/*
+	 * master_drop_all_shards is typically called from the DROP TABLE trigger,
+	 * but could be called by a user directly. Make sure we have an
+	 * AccessExlusiveLock to prevent any other commands from running on this table
+	 * concurrently.
+	 */
+	LockRelationOid(relationId, AccessExclusiveLock);
 
 	shardIntervalList = LoadShardIntervalList(relationId);
 	droppedShardCount = DropShards(relationId, schemaName, relationName,
@@ -237,24 +236,29 @@ master_drop_all_shards(PG_FUNCTION_ARGS)
 
 
 /*
- * master_drop_sequences attempts to drop a list of sequences on a specified
- * node. The "IF EXISTS" clause is used to permit dropping sequences even if
- * they may not exist. Returns true on success, false on failure.
+ * master_drop_sequences attempts to drop a list of sequences on worker nodes.
+ * The "IF EXISTS" clause is used to permit dropping sequences even if they may not
+ * exist. If the commands fail on the workers, the operation is rolled back.
+ * If ddl propagation (citus.enable_ddl_propagation) is set to off, then the function
+ * returns without doing anything.
  */
 Datum
 master_drop_sequences(PG_FUNCTION_ARGS)
 {
 	ArrayType *sequenceNamesArray = PG_GETARG_ARRAYTYPE_P(0);
-	text *nodeText = PG_GETARG_TEXT_P(1);
-	int64 nodePort = PG_GETARG_INT64(2);
-	bool dropSuccessful = false;
-	char *nodeName = TextDatumGetCString(nodeText);
-
 	ArrayIterator sequenceIterator = NULL;
 	Datum sequenceText = 0;
 	bool isNull = false;
-
 	StringInfo dropSeqCommand = makeStringInfo();
+	bool coordinator = IsCoordinator();
+
+	CheckCitusVersion(ERROR);
+
+	/* do nothing if DDL propagation is switched off or this is not the coordinator */
+	if (!EnableDDLPropagation || !coordinator)
+	{
+		PG_RETURN_VOID();
+	}
 
 	/* iterate over sequence names to build single command to DROP them all */
 	sequenceIterator = array_create_iterator(sequenceNamesArray, 0, NULL);
@@ -280,14 +284,44 @@ master_drop_sequences(PG_FUNCTION_ARGS)
 		appendStringInfo(dropSeqCommand, " %s", TextDatumGetCString(sequenceText));
 	}
 
-	dropSuccessful = ExecuteRemoteCommand(nodeName, nodePort, dropSeqCommand);
-	if (!dropSuccessful)
+	if (dropSeqCommand->len != 0)
 	{
-		ereport(WARNING, (errmsg("could not delete sequences from node \"%s:" INT64_FORMAT
-								 "\"", nodeName, nodePort)));
+		appendStringInfoString(dropSeqCommand, " CASCADE");
+
+		SendCommandToWorkers(ALL_WORKERS, DISABLE_DDL_PROPAGATION);
+		SendCommandToWorkers(ALL_WORKERS, dropSeqCommand->data);
 	}
 
-	PG_RETURN_BOOL(dropSuccessful);
+	PG_RETURN_VOID();
+}
+
+
+/*
+ * CheckTableSchemaNameForDrop errors out if the current user does not
+ * have permission to undistribute the given relation, taking into
+ * account that it may be called from the drop trigger. If the table exists,
+ * the function rewrites the given table and schema name.
+ */
+void
+CheckTableSchemaNameForDrop(Oid relationId, char **schemaName, char **tableName)
+{
+	char *tempTableName = get_rel_name(relationId);
+
+	if (tempTableName != NULL)
+	{
+		/* ensure proper values are used if the table exists */
+		Oid schemaId = get_rel_namespace(relationId);
+		(*schemaName) = get_namespace_name(schemaId);
+		(*tableName) = tempTableName;
+
+		EnsureTableOwner(relationId);
+	}
+	else if (!superuser())
+	{
+		/* table does not exist, must be called from drop trigger */
+		ereport(ERROR, (errmsg("cannot drop distributed table metadata as a "
+							   "non-superuser")));
+	}
 }
 
 
@@ -306,14 +340,26 @@ DropShards(Oid relationId, char *schemaName, char *relationName,
 	ListCell *shardIntervalCell = NULL;
 	int droppedShardCount = 0;
 
+	if (XactModificationLevel != XACT_MODIFICATION_NONE)
+	{
+		ereport(ERROR, (errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
+						errmsg("shard drop operations must not appear in "
+							   "transaction blocks containing other distributed "
+							   "modifications")));
+	}
+
+	BeginOrContinueCoordinatedTransaction();
+
+	/* At this point we intentionally decided to not use 2PC for reference tables */
+	if (MultiShardCommitProtocol == COMMIT_PROTOCOL_2PC)
+	{
+		CoordinatedTransactionUse2PC();
+	}
+
 	foreach(shardIntervalCell, deletableShardIntervalList)
 	{
 		List *shardPlacementList = NIL;
-		List *droppedPlacementList = NIL;
-		List *lingeringPlacementList = NIL;
 		ListCell *shardPlacementCell = NULL;
-		ListCell *droppedPlacementCell = NULL;
-		ListCell *lingeringPlacementCell = NULL;
 		ShardInterval *shardInterval = (ShardInterval *) lfirst(shardIntervalCell);
 		uint64 shardId = shardInterval->shardId;
 		char *quotedShardName = NULL;
@@ -332,8 +378,10 @@ DropShards(Oid relationId, char *schemaName, char *relationName,
 				(ShardPlacement *) lfirst(shardPlacementCell);
 			char *workerName = shardPlacement->nodeName;
 			uint32 workerPort = shardPlacement->nodePort;
-			bool dropSuccessful = false;
 			StringInfo workerDropQuery = makeStringInfo();
+			MultiConnection *connection = NULL;
+			uint32 connectionFlags = FOR_DDL;
+			char *extensionOwner = CitusExtensionOwnerName();
 
 			char storageType = shardInterval->storageType;
 			if (storageType == SHARD_STORAGE_TABLE)
@@ -348,58 +396,34 @@ DropShards(Oid relationId, char *schemaName, char *relationName,
 								 quotedShardName);
 			}
 
-			dropSuccessful = ExecuteRemoteCommand(workerName, workerPort,
-												  workerDropQuery);
-			if (dropSuccessful)
-			{
-				droppedPlacementList = lappend(droppedPlacementList, shardPlacement);
-			}
-			else
-			{
-				lingeringPlacementList = lappend(lingeringPlacementList, shardPlacement);
-			}
-		}
+			connection = GetNodeUserDatabaseConnection(connectionFlags, workerName,
+													   workerPort, extensionOwner, NULL);
 
-		/* make sure we don't process cancel signals */
-		HOLD_INTERRUPTS();
+			RemoteTransactionBeginIfNecessary(connection);
 
-		foreach(droppedPlacementCell, droppedPlacementList)
-		{
-			ShardPlacement *placement = (ShardPlacement *) lfirst(droppedPlacementCell);
-			char *workerName = placement->nodeName;
-			uint32 workerPort = placement->nodePort;
+			if (PQstatus(connection->pgConn) != CONNECTION_OK)
+			{
+				uint64 placementId = shardPlacement->placementId;
+
+				ereport(WARNING, (errmsg("could not connect to shard \"%s\" on node "
+										 "\"%s:%u\"", shardRelationName, workerName,
+										 workerPort),
+								  errdetail("Marking this shard placement for "
+											"deletion")));
+
+				UpdateShardPlacementState(placementId, FILE_TO_DELETE);
+
+				continue;
+			}
+
+			MarkRemoteTransactionCritical(connection);
+
+			ExecuteCriticalRemoteCommand(connection, workerDropQuery->data);
 
 			DeleteShardPlacementRow(shardId, workerName, workerPort);
 		}
 
-		/* mark shard placements that we couldn't drop as to be deleted */
-		foreach(lingeringPlacementCell, lingeringPlacementList)
-		{
-			ShardPlacement *placement = (ShardPlacement *) lfirst(lingeringPlacementCell);
-			uint64 placementId = placement->placementId;
-			char *workerName = placement->nodeName;
-			uint32 workerPort = placement->nodePort;
-			uint64 oldShardLength = placement->shardLength;
-
-			DeleteShardPlacementRow(shardId, workerName, workerPort);
-			InsertShardPlacementRow(shardId, placementId, FILE_TO_DELETE, oldShardLength,
-									workerName, workerPort);
-
-			ereport(WARNING, (errmsg("could not delete shard \"%s\" on node \"%s:%u\"",
-									 shardRelationName, workerName, workerPort),
-							  errdetail("Marking this shard placement for deletion")));
-		}
-
 		DeleteShardRow(shardId);
-
-		if (QueryCancelPending)
-		{
-			ereport(WARNING, (errmsg("cancel requests are ignored during shard "
-									 "deletion")));
-			QueryCancelPending = false;
-		}
-
-		RESUME_INTERRUPTS();
 	}
 
 	droppedShardCount = list_length(deletableShardIntervalList);
