@@ -26,8 +26,6 @@
 #include "commands/event_trigger.h"
 #include "distributed/citus_clauses.h"
 #include "distributed/citus_ruleutils.h"
-#include "distributed/commit_protocol.h"
-#include "distributed/connection_cache.h"
 #include "distributed/listutils.h"
 #include "distributed/master_metadata_utility.h"
 #include "distributed/master_protocol.h"
@@ -42,6 +40,7 @@
 #include "distributed/pg_dist_partition.h"
 #include "distributed/resource_lock.h"
 #include "distributed/shardinterval_utils.h"
+#include "distributed/shard_pruning.h"
 #include "distributed/worker_protocol.h"
 #include "optimizer/clauses.h"
 #include "optimizer/predtest.h"
@@ -83,11 +82,14 @@ master_modify_multiple_shards(PG_FUNCTION_ARGS)
 	Node *queryTreeNode;
 	List *restrictClauseList = NIL;
 	bool failOK = false;
-	List *shardIntervalList = NIL;
 	List *prunedShardIntervalList = NIL;
 	List *taskList = NIL;
 	int32 affectedTupleCount = 0;
 
+	EnsureCoordinator();
+	CheckCitusVersion(ERROR);
+
+
 	queryTreeNode = ParseTreeNode(queryString);
 	if (IsA(queryTreeNode, DeleteStmt))
 	{
@@ -139,7 +141,11 @@ master_modify_multiple_shards(PG_FUNCTION_ARGS)
 
 	if (modifyQuery->commandType != CMD_UTILITY)
 	{
-		ErrorIfModifyQueryNotSupported(modifyQuery);
+		DeferredErrorMessage *error = ModifyQuerySupported(modifyQuery);
+		if (error)
+		{
+			RaiseDeferredError(error, ERROR);
+		}
 	}
 
 	/* reject queries with a returning list */
@@ -150,13 +156,12 @@ master_modify_multiple_shards(PG_FUNCTION_ARGS)
 				 errmsg("master_modify_multiple_shards() does not support RETURNING")));
 	}
 
-	ExecuteMasterEvaluableFunctions(modifyQuery);
+	ExecuteMasterEvaluableFunctions(modifyQuery, NULL);
 
-	shardIntervalList = LoadShardIntervalList(relationId);
 	restrictClauseList = WhereClauseList(modifyQuery->jointree);
 
 	prunedShardIntervalList =
-		PruneShardList(relationId, tableId, restrictClauseList, shardIntervalList);
+		PruneShards(relationId, tableId, restrictClauseList);
 
 	CHECK_FOR_INTERRUPTS();
 
@@ -199,6 +204,7 @@ ModifyMultipleShardsTaskList(Query *query, List *shardIntervalList, Oid relation
 		task->taskType = SQL_TASK;
 		task->queryString = shardQueryString->data;
 		task->dependedTaskList = NULL;
+		task->replicationModel = REPLICATION_MODEL_INVALID;
 		task->anchorShardId = shardId;
 		task->taskPlacementList = FinalizedShardPlacementList(shardId);
 
