@@ -26,6 +26,7 @@
 #include "access/htup_details.h"
 #include "access/skey.h"
 #include "access/stratnum.h"
+#include "access/sysattr.h"
 #include "access/tupdesc.h"
 #include "catalog/dependency.h"
 #include "catalog/indexing.h"
@@ -37,11 +38,13 @@
 #endif
 #include "catalog/pg_index.h"
 #include "catalog/pg_type.h"
+#include "catalog/pg_namespace.h"
 #include "commands/sequence.h"
 #include "distributed/citus_ruleutils.h"
 #include "distributed/listutils.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
+#include "distributed/metadata_sync.h"
 #include "distributed/pg_dist_shard.h"
 #include "distributed/worker_manager.h"
 #include "foreign/foreign.h"
@@ -56,11 +59,12 @@
 #include "utils/palloc.h"
 #include "utils/relcache.h"
 #include "utils/ruleutils.h"
+#include "utils/tqual.h"
 
 
 /* Shard related configuration */
 int ShardCount = 32;
-int ShardReplicationFactor = 2; /* desired replication factor for shards */
+int ShardReplicationFactor = 1; /* desired replication factor for shards */
 int ShardMaxSize = 1048576;     /* maximum size in KB one shard can grow to */
 int ShardPlacementPolicy = SHARD_PLACEMENT_ROUND_ROBIN;
 
@@ -73,8 +77,6 @@ PG_FUNCTION_INFO_V1(master_get_table_metadata);
 PG_FUNCTION_INFO_V1(master_get_table_ddl_events);
 PG_FUNCTION_INFO_V1(master_get_new_shardid);
 PG_FUNCTION_INFO_V1(master_get_new_placementid);
-PG_FUNCTION_INFO_V1(master_get_local_first_candidate_nodes);
-PG_FUNCTION_INFO_V1(master_get_round_robin_candidate_nodes);
 PG_FUNCTION_INFO_V1(master_get_active_worker_nodes);
 
 
@@ -91,6 +93,7 @@ master_get_table_metadata(PG_FUNCTION_ARGS)
 	Oid relationId = ResolveRelationId(relationName);
 
 	DistTableCacheEntry *partitionEntry = NULL;
+	char *partitionKeyString = NULL;
 	TypeFuncClass resultTypeClass = 0;
 	Datum partitionKeyExpr = 0;
 	Datum partitionKey = 0;
@@ -102,6 +105,8 @@ master_get_table_metadata(PG_FUNCTION_ARGS)
 	Datum values[TABLE_METADATA_FIELDS];
 	bool isNulls[TABLE_METADATA_FIELDS];
 
+	CheckCitusVersion(ERROR);
+
 	/* find partition tuple for partitioned relation */
 	partitionEntry = DistributedTableCacheEntry(relationId);
 
@@ -112,16 +117,27 @@ master_get_table_metadata(PG_FUNCTION_ARGS)
 		ereport(ERROR, (errmsg("return type must be a row type")));
 	}
 
-	/* get decompiled expression tree for partition key */
-	partitionKeyExpr =
-		PointerGetDatum(cstring_to_text(partitionEntry->partitionKeyString));
-	partitionKey = DirectFunctionCall2(pg_get_expr, partitionKeyExpr,
-									   ObjectIdGetDatum(relationId));
-
 	/* form heap tuple for table metadata */
 	memset(values, 0, sizeof(values));
 	memset(isNulls, false, sizeof(isNulls));
 
+	partitionKeyString = partitionEntry->partitionKeyString;
+
+	/* reference tables do not have partition key */
+	if (partitionKeyString == NULL)
+	{
+		partitionKey = PointerGetDatum(NULL);
+		isNulls[3] = true;
+	}
+	else
+	{
+		/* get decompiled expression tree for partition key */
+		partitionKeyExpr =
+			PointerGetDatum(cstring_to_text(partitionEntry->partitionKeyString));
+		partitionKey = DirectFunctionCall2(pg_get_expr, partitionKeyExpr,
+										   ObjectIdGetDatum(relationId));
+	}
+
 	shardMaxSizeInBytes = (int64) ShardMaxSize * 1024L;
 
 	/* get storage type */
@@ -180,6 +196,8 @@ master_get_table_ddl_events(PG_FUNCTION_ARGS)
 	FuncCallContext *functionContext = NULL;
 	ListCell *tableDDLEventCell = NULL;
 
+	CheckCitusVersion(ERROR);
+
 	/*
 	 * On the very first call to this function, we first use the given relation
 	 * name to get to the relation. We then recreate the list of DDL statements
@@ -190,6 +208,7 @@ master_get_table_ddl_events(PG_FUNCTION_ARGS)
 	{
 		text *relationName = PG_GETARG_TEXT_P(0);
 		Oid relationId = ResolveRelationId(relationName);
+		bool includeSequenceDefaults = true;
 
 		MemoryContext oldContext = NULL;
 		List *tableDDLEventList = NIL;
@@ -201,7 +220,7 @@ master_get_table_ddl_events(PG_FUNCTION_ARGS)
 		oldContext = MemoryContextSwitchTo(functionContext->multi_call_memory_ctx);
 
 		/* allocate DDL statements, and then save position in DDL statements */
-		tableDDLEventList = GetTableDDLEvents(relationId);
+		tableDDLEventList = GetTableDDLEvents(relationId, includeSequenceDefaults);
 		tableDDLEventCell = list_head(tableDDLEventList);
 
 		functionContext->user_fctx = tableDDLEventCell;
@@ -245,8 +264,14 @@ master_get_table_ddl_events(PG_FUNCTION_ARGS)
 Datum
 master_get_new_shardid(PG_FUNCTION_ARGS)
 {
-	uint64 shardId = GetNextShardId();
-	Datum shardIdDatum = Int64GetDatum(shardId);
+	uint64 shardId = 0;
+	Datum shardIdDatum = 0;
+
+	EnsureCoordinator();
+	CheckCitusVersion(ERROR);
+
+	shardId = GetNextShardId();
+	shardIdDatum = Int64GetDatum(shardId);
 
 	PG_RETURN_DATUM(shardIdDatum);
 }
@@ -286,10 +311,9 @@ GetNextShardId()
 
 
 /*
- * master_get_new_placementid allocates and returns a unique placementId for
- * the placement to be created. This allocation occurs both in shared memory
- * and in write ahead logs; writing to logs avoids the risk of having shardId
- * collisions.
+ * master_get_new_placementid is a user facing wrapper function around
+ * GetNextPlacementId() which allocates and returns a unique placement id for the
+ * placement to be created.
  *
  * NB: This can be called by any user; for now we have decided that that's
  * ok. We might want to restrict this to users part of a specific role or such
@@ -297,199 +321,52 @@ GetNextShardId()
  */
 Datum
 master_get_new_placementid(PG_FUNCTION_ARGS)
+{
+	uint64 placementId = 0;
+	Datum placementIdDatum = 0;
+
+	EnsureCoordinator();
+	CheckCitusVersion(ERROR);
+
+	placementId = GetNextPlacementId();
+	placementIdDatum = Int64GetDatum(placementId);
+
+	PG_RETURN_DATUM(placementIdDatum);
+}
+
+
+/*
+ * GetNextPlacementId allocates and returns a unique placementId for
+ * the placement to be created. This allocation occurs both in shared memory
+ * and in write ahead logs; writing to logs avoids the risk of having shardId
+ * collisions.
+ *
+ * NB: This can be called by any user; for now we have decided that that's
+ * ok. We might want to restrict this to users part of a specific role or such
+ * at some later point.
+ */
+uint64
+GetNextPlacementId(void)
 {
 	text *sequenceName = cstring_to_text(PLACEMENTID_SEQUENCE_NAME);
 	Oid sequenceId = ResolveRelationId(sequenceName);
 	Datum sequenceIdDatum = ObjectIdGetDatum(sequenceId);
 	Oid savedUserId = InvalidOid;
 	int savedSecurityContext = 0;
-	Datum shardIdDatum = 0;
+	Datum placementIdDatum = 0;
+	uint64 placementId = 0;
 
 	GetUserIdAndSecContext(&savedUserId, &savedSecurityContext);
 	SetUserIdAndSecContext(CitusExtensionOwner(), SECURITY_LOCAL_USERID_CHANGE);
 
-	/* generate new and unique shardId from sequence */
-	shardIdDatum = DirectFunctionCall1(nextval_oid, sequenceIdDatum);
+	/* generate new and unique placement id from sequence */
+	placementIdDatum = DirectFunctionCall1(nextval_oid, sequenceIdDatum);
 
 	SetUserIdAndSecContext(savedUserId, savedSecurityContext);
 
-	PG_RETURN_DATUM(shardIdDatum);
-}
+	placementId = DatumGetInt64(placementIdDatum);
 
-
-/*
- * master_get_local_first_candidate_nodes returns a set of candidate host names
- * and port numbers on which to place new shards. The function makes sure to
- * always allocate the first candidate node as the node the caller is connecting
- * from; and allocates additional nodes until the shard replication factor is
- * met. The function errors if the caller's remote node name is not found in the
- * membership list, or if the number of available nodes falls short of the
- * replication factor.
- */
-Datum
-master_get_local_first_candidate_nodes(PG_FUNCTION_ARGS)
-{
-	FuncCallContext *functionContext = NULL;
-	uint32 desiredNodeCount = 0;
-	uint32 currentNodeCount = 0;
-
-	if (SRF_IS_FIRSTCALL())
-	{
-		MemoryContext oldContext = NULL;
-		TupleDesc tupleDescriptor = NULL;
-		uint32 liveNodeCount = 0;
-		bool hasOid = false;
-
-		/* create a function context for cross-call persistence */
-		functionContext = SRF_FIRSTCALL_INIT();
-
-		/* switch to memory context appropriate for multiple function calls */
-		oldContext = MemoryContextSwitchTo(functionContext->multi_call_memory_ctx);
-
-		functionContext->user_fctx = NIL;
-		functionContext->max_calls = ShardReplicationFactor;
-
-		/* if enough live nodes, return an extra candidate node as backup */
-		liveNodeCount = WorkerGetLiveNodeCount();
-		if (liveNodeCount > ShardReplicationFactor)
-		{
-			functionContext->max_calls = ShardReplicationFactor + 1;
-		}
-
-		/*
-		 * This tuple descriptor must match the output parameters declared for
-		 * the function in pg_proc.
-		 */
-		tupleDescriptor = CreateTemplateTupleDesc(CANDIDATE_NODE_FIELDS, hasOid);
-		TupleDescInitEntry(tupleDescriptor, (AttrNumber) 1, "node_name",
-						   TEXTOID, -1, 0);
-		TupleDescInitEntry(tupleDescriptor, (AttrNumber) 2, "node_port",
-						   INT8OID, -1, 0);
-
-		functionContext->tuple_desc = BlessTupleDesc(tupleDescriptor);
-
-		MemoryContextSwitchTo(oldContext);
-	}
-
-	functionContext = SRF_PERCALL_SETUP();
-	desiredNodeCount = functionContext->max_calls;
-	currentNodeCount = functionContext->call_cntr;
-
-	if (currentNodeCount < desiredNodeCount)
-	{
-		MemoryContext oldContext = NULL;
-		List *currentNodeList = NIL;
-		WorkerNode *candidateNode = NULL;
-		Datum candidateDatum = 0;
-
-		/* switch to memory context appropriate for multiple function calls */
-		oldContext = MemoryContextSwitchTo(functionContext->multi_call_memory_ctx);
-		currentNodeList = functionContext->user_fctx;
-
-		candidateNode = WorkerGetLocalFirstCandidateNode(currentNodeList);
-		if (candidateNode == NULL)
-		{
-			ereport(ERROR, (errmsg("could only find %u of %u required nodes",
-								   currentNodeCount, desiredNodeCount)));
-		}
-
-		currentNodeList = lappend(currentNodeList, candidateNode);
-		functionContext->user_fctx = currentNodeList;
-
-		MemoryContextSwitchTo(oldContext);
-
-		candidateDatum = WorkerNodeGetDatum(candidateNode, functionContext->tuple_desc);
-
-		SRF_RETURN_NEXT(functionContext, candidateDatum);
-	}
-	else
-	{
-		SRF_RETURN_DONE(functionContext);
-	}
-}
-
-
-/*
- * master_get_round_robin_candidate_nodes returns a set of candidate host names
- * and port numbers on which to place new shards. The function uses the round
- * robin policy to choose the nodes and tries to ensure that there is an even
- * distribution of shards across the worker nodes. This function errors out if
- * the number of available nodes falls short of the replication factor.
- */
-Datum
-master_get_round_robin_candidate_nodes(PG_FUNCTION_ARGS)
-{
-	uint64 shardId = PG_GETARG_INT64(0);
-	FuncCallContext *functionContext = NULL;
-	uint32 desiredNodeCount = 0;
-	uint32 currentNodeCount = 0;
-
-	if (SRF_IS_FIRSTCALL())
-	{
-		MemoryContext oldContext = NULL;
-		TupleDesc tupleDescriptor = NULL;
-		List *workerNodeList = NIL;
-		TypeFuncClass resultTypeClass = 0;
-		uint32 workerNodeCount = 0;
-
-		/* create a function context for cross-call persistence */
-		functionContext = SRF_FIRSTCALL_INIT();
-
-		/* switch to memory context appropriate for multiple function calls */
-		oldContext = MemoryContextSwitchTo(functionContext->multi_call_memory_ctx);
-
-		/* get the worker node list and sort it for determinism */
-		workerNodeList = WorkerNodeList();
-		workerNodeList = SortList(workerNodeList, CompareWorkerNodes);
-
-		functionContext->user_fctx = workerNodeList;
-		functionContext->max_calls = ShardReplicationFactor;
-
-		/* if we enough live nodes, return an extra candidate node as backup */
-		workerNodeCount = (uint32) list_length(workerNodeList);
-		if (workerNodeCount > ShardReplicationFactor)
-		{
-			functionContext->max_calls = ShardReplicationFactor + 1;
-		}
-
-		/* create tuple descriptor for return value */
-		resultTypeClass = get_call_result_type(fcinfo, NULL, &tupleDescriptor);
-		if (resultTypeClass != TYPEFUNC_COMPOSITE)
-		{
-			ereport(ERROR, (errmsg("return type must be a row type")));
-		}
-
-		functionContext->tuple_desc = tupleDescriptor;
-
-		MemoryContextSwitchTo(oldContext);
-	}
-
-	functionContext = SRF_PERCALL_SETUP();
-	desiredNodeCount = functionContext->max_calls;
-	currentNodeCount = functionContext->call_cntr;
-
-	if (currentNodeCount < desiredNodeCount)
-	{
-		List *workerNodeList = functionContext->user_fctx;
-		WorkerNode *candidateNode = NULL;
-		Datum candidateDatum = 0;
-
-		candidateNode = WorkerGetRoundRobinCandidateNode(workerNodeList, shardId,
-														 currentNodeCount);
-		if (candidateNode == NULL)
-		{
-			ereport(ERROR, (errmsg("could only find %u of %u required nodes",
-								   currentNodeCount, desiredNodeCount)));
-		}
-
-		candidateDatum = WorkerNodeGetDatum(candidateNode, functionContext->tuple_desc);
-
-		SRF_RETURN_NEXT(functionContext, candidateDatum);
-	}
-	else
-	{
-		SRF_RETURN_DONE(functionContext);
-	}
+	return placementId;
 }
 
 
@@ -505,6 +382,8 @@ master_get_active_worker_nodes(PG_FUNCTION_ARGS)
 	uint32 workerNodeIndex = 0;
 	uint32 workerNodeCount = 0;
 
+	CheckCitusVersion(ERROR);
+
 	if (SRF_IS_FIRSTCALL())
 	{
 		MemoryContext oldContext = NULL;
@@ -519,7 +398,7 @@ master_get_active_worker_nodes(PG_FUNCTION_ARGS)
 		/* switch to memory context appropriate for multiple function calls */
 		oldContext = MemoryContextSwitchTo(functionContext->multi_call_memory_ctx);
 
-		workerNodeList = WorkerNodeList();
+		workerNodeList = ActiveWorkerNodeList();
 		workerNodeCount = (uint32) list_length(workerNodeList);
 
 		functionContext->user_fctx = workerNodeList;
@@ -580,13 +459,39 @@ ResolveRelationId(text *relationName)
 
 
 /*
- * GetTableDDLEvents takes in a relationId, and returns the list of DDL commands
- * needed to reconstruct the relation. These DDL commands are all palloced; and
- * include the table's schema definition, optional column storage and statistics
- * definitions, and index and constraint definitions.
+ * GetTableDDLEvents takes in a relationId, includeSequenceDefaults flag,
+ * and returns the list of DDL commands needed to reconstruct the relation.
+ * When the flag includeSequenceDefaults is set, the function also creates
+ * DEFAULT clauses for columns getting their default values from a sequence.
+ * These DDL commands are all palloced; and include the table's schema
+ * definition, optional column storage and statistics definitions, and index
+ * and constraint definitions.
  */
 List *
-GetTableDDLEvents(Oid relationId)
+GetTableDDLEvents(Oid relationId, bool includeSequenceDefaults)
+{
+	List *tableDDLEventList = NIL;
+	List *tableCreationCommandList = NIL;
+	List *indexAndConstraintCommandList = NIL;
+
+	tableCreationCommandList = GetTableCreationCommands(relationId,
+														includeSequenceDefaults);
+	tableDDLEventList = list_concat(tableDDLEventList, tableCreationCommandList);
+
+	indexAndConstraintCommandList = GetTableIndexAndConstraintCommands(relationId);
+	tableDDLEventList = list_concat(tableDDLEventList, indexAndConstraintCommandList);
+
+	return tableDDLEventList;
+}
+
+
+/*
+ * GetTableCreationCommands takes in a relationId, and returns the list of DDL
+ * commands needed to reconstruct the relation, excluding indexes and
+ * constraints.
+ */
+List *
+GetTableCreationCommands(Oid relationId, bool includeSequenceDefaults)
 {
 	List *tableDDLEventList = NIL;
 	char tableType = 0;
@@ -594,15 +499,9 @@ GetTableDDLEvents(Oid relationId)
 	ListCell *sequenceIdCell;
 	char *tableSchemaDef = NULL;
 	char *tableColumnOptionsDef = NULL;
-	char *schemaName = NULL;
+	char *createSchemaCommand = NULL;
 	Oid schemaId = InvalidOid;
 
-	Relation pgIndex = NULL;
-	SysScanDesc scanDescriptor = NULL;
-	ScanKeyData scanKey[1];
-	int scanKeyCount = 1;
-	HeapTuple heapTuple = NULL;
-
 	/*
 	 * Set search_path to NIL so that all objects outside of pg_catalog will be
 	 * schema-prefixed. pg_catalog will be added automatically when we call
@@ -629,13 +528,10 @@ GetTableDDLEvents(Oid relationId)
 
 	/* create schema if the table is not in the default namespace (public) */
 	schemaId = get_rel_namespace(relationId);
-	schemaName = get_namespace_name(schemaId);
-	if (strncmp(schemaName, "public", NAMEDATALEN) != 0)
+	createSchemaCommand = CreateSchemaDDLCommand(schemaId);
+	if (createSchemaCommand != NULL)
 	{
-		StringInfo schemaNameDef = makeStringInfo();
-		appendStringInfo(schemaNameDef, CREATE_SCHEMA_COMMAND, schemaName);
-
-		tableDDLEventList = lappend(tableDDLEventList, schemaNameDef->data);
+		tableDDLEventList = lappend(tableDDLEventList, createSchemaCommand);
 	}
 
 	/* create sequences if needed */
@@ -648,7 +544,7 @@ GetTableDDLEvents(Oid relationId)
 	}
 
 	/* fetch table schema and column option definitions */
-	tableSchemaDef = pg_get_tableschemadef_string(relationId);
+	tableSchemaDef = pg_get_tableschemadef_string(relationId, includeSequenceDefaults);
 	tableColumnOptionsDef = pg_get_tablecolumnoptionsdef_string(relationId);
 
 	tableDDLEventList = lappend(tableDDLEventList, tableSchemaDef);
@@ -657,6 +553,37 @@ GetTableDDLEvents(Oid relationId)
 		tableDDLEventList = lappend(tableDDLEventList, tableColumnOptionsDef);
 	}
 
+	/* revert back to original search_path */
+	PopOverrideSearchPath();
+
+	return tableDDLEventList;
+}
+
+
+/*
+ * GetTableIndexAndConstraintCommands returns the list of DDL commands to
+ * (re)create indexes and constraints for a given table.
+ */
+List *
+GetTableIndexAndConstraintCommands(Oid relationId)
+{
+	List *indexDDLEventList = NIL;
+	Relation pgIndex = NULL;
+	SysScanDesc scanDescriptor = NULL;
+	ScanKeyData scanKey[1];
+	int scanKeyCount = 1;
+	HeapTuple heapTuple = NULL;
+
+	/*
+	 * Set search_path to NIL so that all objects outside of pg_catalog will be
+	 * schema-prefixed. pg_catalog will be added automatically when we call
+	 * PushOverrideSearchPath(), since we set addCatalog to true;
+	 */
+	OverrideSearchPath *overridePath = GetOverrideSearchPath(CurrentMemoryContext);
+	overridePath->schemas = NIL;
+	overridePath->addCatalog = true;
+	PushOverrideSearchPath(overridePath);
+
 	/* open system catalog and scan all indexes that belong to this table */
 	pgIndex = heap_open(IndexRelationId, AccessShareLock);
 
@@ -709,7 +636,7 @@ GetTableDDLEvents(Oid relationId)
 		}
 
 		/* append found constraint or index definition to the list */
-		tableDDLEventList = lappend(tableDDLEventList, statementDef);
+		indexDDLEventList = lappend(indexDDLEventList, statementDef);
 
 		/* if table is clustered on this index, append definition to the list */
 		if (indexForm->indisclustered)
@@ -717,7 +644,7 @@ GetTableDDLEvents(Oid relationId)
 			char *clusteredDef = pg_get_indexclusterdef_string(indexId);
 			Assert(clusteredDef != NULL);
 
-			tableDDLEventList = lappend(tableDDLEventList, clusteredDef);
+			indexDDLEventList = lappend(indexDDLEventList, clusteredDef);
 		}
 
 		heapTuple = systable_getnext(scanDescriptor);
@@ -730,7 +657,7 @@ GetTableDDLEvents(Oid relationId)
 	/* revert back to original search_path */
 	PopOverrideSearchPath();
 
-	return tableDDLEventList;
+	return indexDDLEventList;
 }
 
 
@@ -830,6 +757,17 @@ ShardStorageType(Oid relationId)
 }
 
 
+/*
+ * IsCoordinator function returns true if this node is identified as the
+ * schema/coordinator/master node of the cluster.
+ */
+bool
+IsCoordinator(void)
+{
+	return (GetLocalGroupId() == 0);
+}
+
+
 /*
  * WorkerNodeGetDatum converts the worker node passed to it into its datum
  * representation. To do this, the function first creates the heap tuple from
