@@ -19,7 +19,7 @@
 
 #include "catalog/pg_class.h"
 #include "distributed/colocation_utils.h"
-#include "distributed/connection_cache.h"
+#include "distributed/connection_management.h"
 #include "distributed/listutils.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
@@ -45,14 +45,8 @@ static void RepairShardPlacement(int64 shardId, char *sourceNodeName,
 static void EnsureShardCanBeRepaired(int64 shardId, char *sourceNodeName,
 									 int32 sourceNodePort, char *targetNodeName,
 									 int32 targetNodePort);
-static ShardPlacement * SearchShardPlacementInList(List *shardPlacementList,
-												   char *nodeName, uint32 nodePort,
-												   bool missingOk);
-static List * CopyShardCommandList(ShardInterval *shardInterval, char *sourceNodeName,
-								   int32 sourceNodePort);
-static List * CopyShardForeignConstraintCommandList(ShardInterval *shardInterval);
-static char * ConstructQualifiedShardName(ShardInterval *shardInterval);
 static List * RecreateTableDDLCommandList(Oid relationId);
+static List * WorkerApplyShardDDLCommandList(List *ddlCommandList, int64 shardId);
 
 /* declarations for dynamic loading */
 PG_FUNCTION_INFO_V1(master_copy_shard_placement);
@@ -90,6 +84,9 @@ master_copy_shard_placement(PG_FUNCTION_ARGS)
 							   "is only supported on Citus Enterprise")));
 	}
 
+	EnsureCoordinator();
+	CheckCitusVersion(ERROR);
+
 	/* RepairShardPlacement function repairs only given shard */
 	RepairShardPlacement(shardId, sourceNodeName, sourceNodePort, targetNodeName,
 						 targetNodePort);
@@ -208,7 +205,7 @@ EnsureShardCanBeRepaired(int64 shardId, char *sourceNodeName, int32 sourceNodePo
  * specified node name and port. If missingOk is set to true, this function returns NULL
  * if no such placement exists in the provided list, otherwise it throws an error.
  */
-static ShardPlacement *
+ShardPlacement *
 SearchShardPlacementInList(List *shardPlacementList, char *nodeName, uint32 nodePort, bool
 						   missingOk)
 {
@@ -248,31 +245,24 @@ SearchShardPlacementInList(List *shardPlacementList, char *nodeName, uint32 node
  * CopyShardCommandList generates command list to copy the given shard placement
  * from the source node to the target node.
  */
-static List *
+List *
 CopyShardCommandList(ShardInterval *shardInterval,
 					 char *sourceNodeName, int32 sourceNodePort)
 {
+	int64 shardId = shardInterval->shardId;
 	char *shardName = ConstructQualifiedShardName(shardInterval);
-	List *ddlCommandList = NIL;
-	ListCell *ddlCommandCell = NULL;
+	List *tableRecreationCommandList = NIL;
+	List *indexCommandList = NIL;
 	List *copyShardToNodeCommandsList = NIL;
 	StringInfo copyShardDataCommand = makeStringInfo();
+	Oid relationId = shardInterval->relationId;
 
-	ddlCommandList = RecreateTableDDLCommandList(shardInterval->relationId);
+	tableRecreationCommandList = RecreateTableDDLCommandList(relationId);
+	tableRecreationCommandList =
+		WorkerApplyShardDDLCommandList(tableRecreationCommandList, shardId);
 
-	foreach(ddlCommandCell, ddlCommandList)
-	{
-		char *ddlCommand = lfirst(ddlCommandCell);
-		char *escapedDdlCommand = quote_literal_cstr(ddlCommand);
-
-		StringInfo applyDdlCommand = makeStringInfo();
-		appendStringInfo(applyDdlCommand,
-						 WORKER_APPLY_SHARD_DDL_COMMAND_WITHOUT_SCHEMA,
-						 shardInterval->shardId, escapedDdlCommand);
-
-		copyShardToNodeCommandsList = lappend(copyShardToNodeCommandsList,
-											  applyDdlCommand->data);
-	}
+	copyShardToNodeCommandsList = list_concat(copyShardToNodeCommandsList,
+											  tableRecreationCommandList);
 
 	appendStringInfo(copyShardDataCommand, WORKER_APPEND_TABLE_TO_SHARD,
 					 quote_literal_cstr(shardName), /* table to append */
@@ -283,6 +273,12 @@ CopyShardCommandList(ShardInterval *shardInterval,
 	copyShardToNodeCommandsList = lappend(copyShardToNodeCommandsList,
 										  copyShardDataCommand->data);
 
+	indexCommandList = GetTableIndexAndConstraintCommands(relationId);
+	indexCommandList = WorkerApplyShardDDLCommandList(indexCommandList, shardId);
+
+	copyShardToNodeCommandsList = list_concat(copyShardToNodeCommandsList,
+											  indexCommandList);
+
 	return copyShardToNodeCommandsList;
 }
 
@@ -291,7 +287,7 @@ CopyShardCommandList(ShardInterval *shardInterval,
  * CopyShardForeignConstraintCommandList generates command list to create foreign
  * constraints existing in source shard after copying it to the other node.
  */
-static List *
+List *
 CopyShardForeignConstraintCommandList(ShardInterval *shardInterval)
 {
 	List *copyShardForeignConstraintCommandList = NIL;
@@ -307,7 +303,7 @@ CopyShardForeignConstraintCommandList(ShardInterval *shardInterval)
 	/* we will only use shardIndex if there is a foreign constraint */
 	if (commandList != NIL)
 	{
-		shardIndex = FindShardIntervalIndex(shardInterval);
+		shardIndex = ShardIndex(shardInterval);
 	}
 
 	foreach(commandCell, commandList)
@@ -354,10 +350,8 @@ CopyShardForeignConstraintCommandList(ShardInterval *shardInterval)
 /*
  * ConstuctQualifiedShardName creates the fully qualified name string of the
  * given shard in <schema>.<table_name>_<shard_id> format.
- *
- * FIXME: Copied from Citus-MX, should be removed once those changes checked-in to Citus.
  */
-static char *
+char *
 ConstructQualifiedShardName(ShardInterval *shardInterval)
 {
 	Oid schemaId = get_rel_namespace(shardInterval->relationId);
@@ -375,8 +369,8 @@ ConstructQualifiedShardName(ShardInterval *shardInterval)
 
 /*
  * RecreateTableDDLCommandList returns a list of DDL statements similar to that
- * returned by GetTableDDLEvents except that the list begins with a "DROP TABLE"
- * or "DROP FOREIGN TABLE" statement to facilitate total recreation of a placement.
+ * returned by GetTableCreationCommands except that the list begins with a "DROP TABLE"
+ * or "DROP FOREIGN TABLE" statement to facilitate idempotent recreation of a placement.
  */
 static List *
 RecreateTableDDLCommandList(Oid relationId)
@@ -392,6 +386,7 @@ RecreateTableDDLCommandList(Oid relationId)
 	List *dropCommandList = NIL;
 	List *recreateCommandList = NIL;
 	char relationKind = get_rel_relkind(relationId);
+	bool includeSequenceDefaults = false;
 
 	/* build appropriate DROP command based on relation kind */
 	if (relationKind == RELKIND_RELATION)
@@ -411,10 +406,36 @@ RecreateTableDDLCommandList(Oid relationId)
 	}
 
 	dropCommandList = list_make1(dropCommand->data);
-
-	createCommandList = GetTableDDLEvents(relationId);
-
+	createCommandList = GetTableCreationCommands(relationId, includeSequenceDefaults);
 	recreateCommandList = list_concat(dropCommandList, createCommandList);
 
 	return recreateCommandList;
 }
+
+
+/*
+ * WorkerApplyShardDDLCommandList wraps all DDL commands in ddlCommandList
+ * in a call to worker_apply_shard_ddl_command to apply the DDL command to
+ * the shard specified by shardId.
+ */
+static List *
+WorkerApplyShardDDLCommandList(List *ddlCommandList, int64 shardId)
+{
+	List *applyDdlCommandList = NIL;
+	ListCell *ddlCommandCell = NULL;
+
+	foreach(ddlCommandCell, ddlCommandList)
+	{
+		char *ddlCommand = lfirst(ddlCommandCell);
+		char *escapedDdlCommand = quote_literal_cstr(ddlCommand);
+
+		StringInfo applyDdlCommand = makeStringInfo();
+		appendStringInfo(applyDdlCommand,
+						 WORKER_APPLY_SHARD_DDL_COMMAND_WITHOUT_SCHEMA,
+						 shardId, escapedDdlCommand);
+
+		applyDdlCommandList = lappend(applyDdlCommandList, applyDdlCommand->data);
+	}
+
+	return applyDdlCommandList;
+}
