@@ -8,6 +8,7 @@
  */
 
 #include "postgres.h"
+#include "libpq-fe.h"
 #include "miscadmin.h"
 
 #include "access/xact.h"
@@ -20,6 +21,7 @@
 #include "commands/tablecmds.h"
 #include "optimizer/cost.h"
 #include "distributed/citus_nodefuncs.h"
+#include "distributed/connection_management.h"
 #include "distributed/multi_client_executor.h"
 #include "distributed/multi_executor.h"
 #include "distributed/multi_explain.h"
@@ -29,6 +31,8 @@
 #include "distributed/multi_physical_planner.h"
 #include "distributed/multi_planner.h"
 #include "distributed/multi_server_executor.h"
+#include "distributed/remote_commands.h"
+#include "distributed/placement_connection.h"
 #include "distributed/worker_protocol.h"
 #include "lib/stringinfo.h"
 #include "nodes/plannodes.h"
@@ -52,8 +56,6 @@
 
 
 /* Config variables that enable printing distributed query plans */
-bool ExplainMultiLogicalPlan = false;
-bool ExplainMultiPhysicalPlan = false;
 bool ExplainDistributedQueries = true;
 bool ExplainAllTasks = false;
 
@@ -67,9 +69,6 @@ typedef struct RemoteExplainPlan
 
 
 /* Explain functions for distributed queries */
-static void ExplainMasterPlan(PlannedStmt *masterPlan, IntoClause *into,
-							  ExplainState *es, const char *queryString,
-							  ParamListInfo params, const instr_time *planDuration);
 static void ExplainJob(Job *job, ExplainState *es);
 static void ExplainMapMergeJob(MapMergeJob *mapMergeJob, ExplainState *es);
 static void ExplainTaskList(List *taskList, ExplainState *es);
@@ -91,325 +90,29 @@ static void ExplainYAMLLineStarting(ExplainState *es);
 
 
 /*
- * MultiExplainOneQuery takes the given query, and checks if the query is local
- * or distributed. If the query is local, the function runs the standard explain
- * logic. If the query is distributed, the function looks up configuration and
- * prints out the distributed logical and physical plans as appropriate.
+ * CitusExplainScan is a custom scan explain callback function which is used to
+ * print explain information of a Citus plan which includes both master and
+ * distributed plan.
  */
 void
-MultiExplainOneQuery(Query *query, IntoClause *into, ExplainState *es,
-					 const char *queryString, ParamListInfo params)
+CitusExplainScan(CustomScanState *node, List *ancestors, struct ExplainState *es)
 {
-	MultiPlan *multiPlan = NULL;
-	CmdType commandType = CMD_UNKNOWN;
-	PlannedStmt *initialPlan = NULL;
-	Job *workerJob = NULL;
-	bool routerExecutablePlan = false;
-	instr_time planStart;
-	instr_time planDuration;
-	Query *originalQuery = NULL;
-	RelationRestrictionContext *restrictionContext = NULL;
-	bool localQuery = !NeedsDistributedPlanning(query);
-	int cursorOptions = 0;
-
-#if PG_VERSION_NUM >= 90600
-
-	/*
-	 * Allow parallel plans in 9.6+ unless selecting into a table.
-	 * Without this, we're breaking explain for non-Citus plans.
-	 */
-	if (!into)
-	{
-		cursorOptions |= CURSOR_OPT_PARALLEL_OK;
-	}
-#endif
-
-	/* handle local queries in the same way as ExplainOneQuery */
-	if (localQuery)
-	{
-		PlannedStmt *plan = NULL;
-
-		INSTR_TIME_SET_CURRENT(planStart);
-
-		/* plan the query */
-		plan = pg_plan_query(query, cursorOptions, params);
-
-		INSTR_TIME_SET_CURRENT(planDuration);
-		INSTR_TIME_SUBTRACT(planDuration, planStart);
-
-		/* run it (if needed) and produce output */
-		ExplainOnePlan(plan, into, es, queryString, params, &planDuration);
-
-		return;
-	}
-
-	/*
-	 * standard_planner scribbles on it's input, but for deparsing we need the
-	 * unmodified form. So copy once we're sure it's a distributed query.
-	 */
-	originalQuery = copyObject(query);
-
-	/* measure the full planning time to display in EXPLAIN ANALYZE */
-	INSTR_TIME_SET_CURRENT(planStart);
-
-	restrictionContext = CreateAndPushRestrictionContext();
-
-	PG_TRY();
-	{
-		/* call standard planner to modify the query structure before multi planning */
-		initialPlan = standard_planner(query, cursorOptions, params);
-
-		commandType = initialPlan->commandType;
-		if (commandType == CMD_INSERT || commandType == CMD_UPDATE ||
-			commandType == CMD_DELETE)
-		{
-			if (es->analyze)
-			{
-				ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-								errmsg("Using ANALYZE for INSERT/UPDATE/DELETE on "
-									   "distributed tables is not supported.")));
-			}
-		}
-
-		multiPlan = CreatePhysicalPlan(originalQuery, query, restrictionContext);
-	}
-	PG_CATCH();
-	{
-		PopRestrictionContext();
-		PG_RE_THROW();
-	}
-	PG_END_TRY();
-
-	PopRestrictionContext();
-
-	INSTR_TIME_SET_CURRENT(planDuration);
-	INSTR_TIME_SUBTRACT(planDuration, planStart);
-
-	if (ExplainMultiLogicalPlan)
-	{
-		MultiTreeRoot *multiTree = MultiLogicalPlanCreate(query);
-		char *logicalPlanString = CitusNodeToString(multiTree);
-		char *formattedPlanString = pretty_format_node_dump(logicalPlanString);
-
-		appendStringInfo(es->str, "logical plan:\n");
-		appendStringInfo(es->str, "%s\n", formattedPlanString);
-	}
-
-	if (ExplainMultiPhysicalPlan)
-	{
-		char *physicalPlanString = CitusNodeToString(multiPlan);
-		char *formattedPlanString = pretty_format_node_dump(physicalPlanString);
-
-		appendStringInfo(es->str, "physical plan:\n");
-		appendStringInfo(es->str, "%s\n", formattedPlanString);
-	}
+	CitusScanState *scanState = (CitusScanState *) node;
+	MultiPlan *multiPlan = scanState->multiPlan;
 
 	if (!ExplainDistributedQueries)
 	{
+		appendStringInfoSpaces(es->str, es->indent * 2);
 		appendStringInfo(es->str, "explain statements for distributed queries ");
 		appendStringInfo(es->str, "are not enabled\n");
 		return;
 	}
 
-	ExplainOpenGroup("Distributed Query", NULL, true, es);
+	ExplainOpenGroup("Distributed Query", "Distributed Query", true, es);
 
-	if (es->format == EXPLAIN_FORMAT_TEXT)
-	{
-		appendStringInfoSpaces(es->str, es->indent * 2);
-		appendStringInfo(es->str, "Distributed Query");
+	ExplainJob(multiPlan->workerJob, es);
 
-		if (multiPlan->masterTableName != NULL)
-		{
-			appendStringInfo(es->str, " into %s", multiPlan->masterTableName);
-		}
-
-		appendStringInfo(es->str, "\n");
-
-		es->indent += 1;
-	}
-
-	routerExecutablePlan = multiPlan->routerExecutable;
-
-	if (routerExecutablePlan)
-	{
-		ExplainPropertyText("Executor", "Router", es);
-	}
-	else
-	{
-		switch (TaskExecutorType)
-		{
-			case MULTI_EXECUTOR_REAL_TIME:
-			{
-				ExplainPropertyText("Executor", "Real-Time", es);
-			}
-			break;
-
-			case MULTI_EXECUTOR_TASK_TRACKER:
-			{
-				ExplainPropertyText("Executor", "Task-Tracker", es);
-			}
-			break;
-
-			default:
-			{
-				ExplainPropertyText("Executor", "Other", es);
-			}
-			break;
-		}
-	}
-
-	workerJob = multiPlan->workerJob;
-	ExplainJob(workerJob, es);
-
-	if (es->format == EXPLAIN_FORMAT_TEXT)
-	{
-		es->indent -= 1;
-	}
-
-	if (!routerExecutablePlan)
-	{
-		PlannedStmt *masterPlan = MultiQueryContainerNode(initialPlan, multiPlan);
-
-		if (es->format == EXPLAIN_FORMAT_TEXT)
-		{
-			appendStringInfoSpaces(es->str, es->indent * 2);
-			appendStringInfo(es->str, "Master Query\n");
-			es->indent += 1;
-		}
-
-		ExplainOpenGroup("Master Query", "Master Query", false, es);
-
-		ExplainMasterPlan(masterPlan, into, es, queryString, params, &planDuration);
-
-		ExplainCloseGroup("Master Query", "Master Query", false, es);
-
-		if (es->format == EXPLAIN_FORMAT_TEXT)
-		{
-			es->indent -= 1;
-		}
-	}
-
-	ExplainCloseGroup("Distributed Query", NULL, true, es);
-}
-
-
-/*
- * ExplainMasterPlan generates EXPLAIN output for the master query that merges results.
- * When using EXPLAIN ANALYZE, this function shows the execution time of the master query
- * in isolation. Calling ExplainOnePlan directly would show the overall execution time of
- * the distributed query, which makes it hard to determine how much time the master query
- * took.
- *
- * Parts of this function are copied directly from ExplainOnePlan.
- */
-static void
-ExplainMasterPlan(PlannedStmt *masterPlan, IntoClause *into,
-				  ExplainState *es, const char *queryString,
-				  ParamListInfo params, const instr_time *planDuration)
-{
-	DestReceiver *dest = NULL;
-	int eflags = 0;
-	QueryDesc *queryDesc = NULL;
-	int instrument_option = 0;
-
-	if (es->analyze && es->timing)
-	{
-		instrument_option |= INSTRUMENT_TIMER;
-	}
-	else if (es->analyze)
-	{
-		instrument_option |= INSTRUMENT_ROWS;
-	}
-
-	if (es->buffers)
-	{
-		instrument_option |= INSTRUMENT_BUFFERS;
-	}
-
-	/*
-	 * Use a snapshot with an updated command ID to ensure this query sees
-	 * results of any previously executed queries.
-	 */
-	PushCopiedSnapshot(GetActiveSnapshot());
-	UpdateActiveSnapshotCommandId();
-
-	/*
-	 * Normally we discard the query's output, but if explaining CREATE TABLE
-	 * AS, we'd better use the appropriate tuple receiver.
-	 */
-	if (into)
-	{
-		dest = CreateIntoRelDestReceiver(into);
-	}
-	else
-	{
-		dest = None_Receiver;
-	}
-
-	/* Create a QueryDesc for the query */
-	queryDesc = CreateQueryDesc(masterPlan, queryString,
-								GetActiveSnapshot(), InvalidSnapshot,
-								dest, params, instrument_option);
-
-	/* Select execution options */
-	if (es->analyze)
-	{
-		eflags = 0;             /* default run-to-completion flags */
-	}
-	else
-	{
-		eflags = EXEC_FLAG_EXPLAIN_ONLY;
-	}
-	if (into)
-	{
-		eflags |= GetIntoRelEFlags(into);
-	}
-
-	/*
-	 * ExecutorStart creates the merge table. If using ANALYZE, it also executes the
-	 * worker job and populates the merge table.
-	 */
-	ExecutorStart(queryDesc, eflags);
-
-	if (es->analyze)
-	{
-		ScanDirection dir;
-
-		/* if using analyze, then finish query execution */
-
-		/* EXPLAIN ANALYZE CREATE TABLE AS WITH NO DATA is weird */
-		if (into && into->skipData)
-		{
-			dir = NoMovementScanDirection;
-		}
-		else
-		{
-			dir = ForwardScanDirection;
-		}
-
-		/* run the plan */
-		ExecutorRun(queryDesc, dir, 0L);
-
-		/* run cleanup too */
-		ExecutorFinish(queryDesc);
-	}
-
-	/*
-	 * ExplainOnePlan executes the master query again, which ensures that the execution
-	 * time only shows the execution time of the master query itself, instead of the
-	 * overall execution time.
-	 */
-	ExplainOnePlan(queryDesc->plannedstmt, into, es, queryString, params, planDuration);
-
-	/*
-	 * ExecutorEnd for the distributed query is deferred until after the master query
-	 * is executed again, otherwise the merge table would be dropped.
-	 */
-	ExecutorEnd(queryDesc);
-
-	FreeQueryDesc(queryDesc);
-
-	PopActiveSnapshot();
+	ExplainCloseGroup("Distributed Query", "Distributed Query", true, es);
 }
 
 
@@ -589,15 +292,60 @@ RemoteExplain(Task *task, ExplainState *es)
 	remotePlan = (RemoteExplainPlan *) palloc0(sizeof(RemoteExplainPlan));
 	explainQuery = BuildRemoteExplainQuery(task->queryString, es);
 
+	/*
+	 * Use a coordinated transaction to ensure that we open a transaction block
+	 * such that we can set a savepoint.
+	 */
+	BeginOrContinueCoordinatedTransaction();
+
 	for (placementIndex = 0; placementIndex < placementCount; placementIndex++)
 	{
 		ShardPlacement *taskPlacement = list_nth(taskPlacementList, placementIndex);
-		char *nodeName = taskPlacement->nodeName;
-		uint32 nodePort = taskPlacement->nodePort;
+		MultiConnection *connection = NULL;
+		PGresult *queryResult = NULL;
+		int connectionFlags = 0;
+		int executeResult = 0;
 
 		remotePlan->placementIndex = placementIndex;
-		remotePlan->explainOutputList = ExecuteRemoteQuery(nodeName, nodePort,
-														   NULL, explainQuery);
+
+		connection = GetPlacementConnection(connectionFlags, taskPlacement, NULL);
+
+		/* try other placements if we fail to connect this one */
+		if (PQstatus(connection->pgConn) != CONNECTION_OK)
+		{
+			continue;
+		}
+
+		RemoteTransactionBeginIfNecessary(connection);
+
+		/*
+		 * Start a savepoint for the explain query. After running the explain
+		 * query, we will rollback to this savepoint. This saves us from side
+		 * effects of EXPLAIN ANALYZE on DML queries.
+		 */
+		ExecuteCriticalRemoteCommand(connection, "SAVEPOINT citus_explain_savepoint");
+
+		/* run explain query */
+		executeResult = ExecuteOptionalRemoteCommand(connection, explainQuery->data,
+													 &queryResult);
+		if (executeResult != 0)
+		{
+			PQclear(queryResult);
+			ForgetResults(connection);
+
+			continue;
+		}
+
+		/* read explain query results */
+		remotePlan->explainOutputList = ReadFirstColumnAsText(queryResult);
+
+		PQclear(queryResult);
+		ForgetResults(connection);
+
+		/* rollback to the savepoint */
+		ExecuteCriticalRemoteCommand(connection,
+									 "ROLLBACK TO SAVEPOINT citus_explain_savepoint");
+
 		if (remotePlan->explainOutputList != NIL)
 		{
 			break;
