@@ -22,8 +22,10 @@
 
 #include "distributed/citus_nodefuncs.h"
 #include "distributed/citus_nodes.h"
+#include "distributed/errormessage.h"
 #include "distributed/multi_logical_planner.h"
 #include "distributed/multi_physical_planner.h"
+#include "distributed/multi_planner.h"
 #include "distributed/master_metadata_utility.h"
 #include "lib/stringinfo.h"
 #include "nodes/plannodes.h"
@@ -274,10 +276,13 @@ OutMultiPlan(OUTFUNC_ARGS)
 
 	WRITE_NODE_TYPE("MULTIPLAN");
 
+	WRITE_INT_FIELD(operation);
+	WRITE_BOOL_FIELD(hasReturning);
+
 	WRITE_NODE_FIELD(workerJob);
 	WRITE_NODE_FIELD(masterQuery);
-	WRITE_STRING_FIELD(masterTableName);
 	WRITE_BOOL_FIELD(routerExecutable);
+	WRITE_NODE_FIELD(planningError);
 }
 
 
@@ -391,6 +396,7 @@ OutJobFields(StringInfo str, const Job *node)
 	WRITE_NODE_FIELD(dependedJobList);
 	WRITE_BOOL_FIELD(subqueryPushdown);
 	WRITE_BOOL_FIELD(requiresMasterEvaluation);
+	WRITE_BOOL_FIELD(deferredPruning);
 }
 
 
@@ -472,6 +478,21 @@ OutShardPlacement(OUTFUNC_ARGS)
 	WRITE_ENUM_FIELD(shardState, RelayFileState);
 	WRITE_STRING_FIELD(nodeName);
 	WRITE_UINT_FIELD(nodePort);
+	/* so we can deal with 0 */
+	WRITE_INT_FIELD(partitionMethod);
+	WRITE_UINT_FIELD(colocationGroupId);
+	WRITE_UINT_FIELD(representativeValue);
+}
+
+
+void
+OutRelationShard(OUTFUNC_ARGS)
+{
+	WRITE_LOCALS(RelationShard);
+	WRITE_NODE_TYPE("RELATIONSHARD");
+
+	WRITE_OID_FIELD(relationId);
+	WRITE_UINT64_FIELD(shardId);
 }
 
 
@@ -494,10 +515,28 @@ OutTask(OUTFUNC_ARGS)
 	WRITE_BOOL_FIELD(assignmentConstrained);
 	WRITE_NODE_FIELD(taskExecution);
 	WRITE_BOOL_FIELD(upsertQuery);
+	WRITE_CHAR_FIELD(replicationModel);
 	WRITE_BOOL_FIELD(insertSelectQuery);
-	WRITE_NODE_FIELD(selectShardList);
+	WRITE_NODE_FIELD(relationShardList);
 }
 
+
+void
+OutDeferredErrorMessage(OUTFUNC_ARGS)
+{
+	WRITE_LOCALS(DeferredErrorMessage);
+	WRITE_NODE_TYPE("DEFERREDERRORMESSAGE");
+
+	WRITE_INT_FIELD(code);
+	WRITE_STRING_FIELD(message);
+	WRITE_STRING_FIELD(detail);
+	WRITE_STRING_FIELD(hint);
+	WRITE_STRING_FIELD(filename);
+	WRITE_INT_FIELD(linenumber);
+	WRITE_STRING_FIELD(functionname);
+}
+
+
 #if (PG_VERSION_NUM < 90600)
 
 /*
@@ -612,6 +651,18 @@ outNode(StringInfo str, const void *obj)
 			appendStringInfoChar(str, '}');
 			break;
 
+		case T_RelationShard:
+			appendStringInfoChar(str, '{');
+			OutRelationShard(str, obj);
+			appendStringInfoChar(str, '}');
+			break;
+
+		case T_DeferredErrorMessage:
+			appendStringInfoChar(str, '{');
+			OutDeferredErrorMessage(str, obj);
+			appendStringInfoChar(str, '}');
+			break;
+
 		default:
 			/* fall back into postgres' normal nodeToString machinery */
 			appendStringInfoString(str, nodeToString(obj));
