@@ -1515,8 +1515,12 @@ CitusParseNodeString(void)
 		return_value = ReadMapMergeJob();
 	else if (MATCH("SHARDPLACEMENT", 14))
 		return_value = ReadShardPlacement();
+	else if (MATCH("RELATIONSHARD", 13))
+		return_value = ReadRelationShard();
 	else if (MATCH("TASK", 4))
 		return_value = ReadTask();
+	else if (MATCH("DEFERREDERRORMESSAGE", 20))
+		return_value = ReadDeferredErrorMessage();
 /* XXX: END Citus Nodes */
 	else
 	{
