@@ -10,20 +10,26 @@
  */
 
 #include "postgres.h"
+#include "miscadmin.h"
 
 #include "access/genam.h"
 #include "access/heapam.h"
 #include "access/htup_details.h"
 #include "access/xact.h"
 #include "catalog/indexing.h"
+#include "commands/sequence.h"
 #include "distributed/colocation_utils.h"
 #include "distributed/listutils.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
+#include "distributed/metadata_sync.h"
 #include "distributed/multi_logical_planner.h"
+#include "distributed/pg_dist_colocation.h"
 #include "distributed/resource_lock.h"
 #include "distributed/shardinterval_utils.h"
 #include "distributed/worker_protocol.h"
+#include "distributed/worker_transaction.h"
+#include "utils/builtins.h"
 #include "utils/fmgroids.h"
 #include "utils/lsyscache.h"
 #include "utils/rel.h"
@@ -34,9 +40,13 @@ static void MarkTablesColocated(Oid sourceRelationId, Oid targetRelationId);
 static void ErrorIfShardPlacementsNotColocated(Oid leftRelationId, Oid rightRelationId);
 static bool ShardsIntervalsEqual(ShardInterval *leftShardInterval,
 								 ShardInterval *rightShardInterval);
+static bool HashPartitionedShardIntervalsEqual(ShardInterval *leftShardInterval,
+											   ShardInterval *rightShardInterval);
 static int CompareShardPlacementsByNode(const void *leftElement,
 										const void *rightElement);
 static void UpdateRelationColocationGroup(Oid distributedRelationId, uint32 colocationId);
+static List * ColocationGroupTableList(Oid colocationId);
+static void DeleteColocationGroup(uint32 colocationId);
 
 
 /* exports for SQL callable functions */
@@ -64,11 +74,15 @@ mark_tables_colocated(PG_FUNCTION_ARGS)
 							   "operation")));
 	}
 
+	EnsureCoordinator();
+	CheckCitusVersion(ERROR);
+
 	relationIdDatumArray = DeconstructArrayObject(relationIdArrayObject);
 
 	for (relationIndex = 0; relationIndex < relationCount; relationIndex++)
 	{
 		Oid nextRelationOid = DatumGetObjectId(relationIdDatumArray[relationIndex]);
+
 		MarkTablesColocated(sourceRelationId, nextRelationOid);
 	}
 
@@ -86,32 +100,11 @@ static void
 MarkTablesColocated(Oid sourceRelationId, Oid targetRelationId)
 {
 	uint32 sourceColocationId = INVALID_COLOCATION_ID;
+	uint32 targetColocationId = INVALID_COLOCATION_ID;
 	Relation pgDistColocation = NULL;
-	Var *sourceDistributionColumn = NULL;
-	Var *targetDistributionColumn = NULL;
-	Oid sourceDistributionColumnType = InvalidOid;
-	Oid targetDistributionColumnType = InvalidOid;
 
-	CheckHashPartitionedTable(sourceRelationId);
-	CheckHashPartitionedTable(targetRelationId);
-
-	sourceDistributionColumn = PartitionKey(sourceRelationId);
-	sourceDistributionColumnType = sourceDistributionColumn->vartype;
-
-	targetDistributionColumn = PartitionKey(targetRelationId);
-	targetDistributionColumnType = targetDistributionColumn->vartype;
-
-	if (sourceDistributionColumnType != targetDistributionColumnType)
-	{
-		char *sourceRelationName = get_rel_name(sourceRelationId);
-		char *targetRelationName = get_rel_name(targetRelationId);
-
-		ereport(ERROR, (errmsg("cannot colocate tables %s and %s",
-							   sourceRelationName, targetRelationName),
-						errdetail("Distribution column types don't match for "
-								  "%s and %s.", sourceRelationName,
-								  targetRelationName)));
-	}
+	CheckReplicationModel(sourceRelationId, targetRelationId);
+	CheckDistributionColumnType(sourceRelationId, targetRelationId);
 
 	/*
 	 * Get an exclusive lock on the colocation system catalog. Therefore, we
@@ -133,14 +126,28 @@ MarkTablesColocated(Oid sourceRelationId, Oid targetRelationId)
 		uint32 shardCount = ShardIntervalCount(sourceRelationId);
 		uint32 shardReplicationFactor = TableShardReplicationFactor(sourceRelationId);
 
+		Var *sourceDistributionColumn = PartitionKey(sourceRelationId);
+		Oid sourceDistributionColumnType = InvalidOid;
+
+		/* reference tables has NULL distribution column */
+		if (sourceDistributionColumn != NULL)
+		{
+			sourceDistributionColumnType = sourceDistributionColumn->vartype;
+		}
+
 		sourceColocationId = CreateColocationGroup(shardCount, shardReplicationFactor,
 												   sourceDistributionColumnType);
 		UpdateRelationColocationGroup(sourceRelationId, sourceColocationId);
 	}
 
+	targetColocationId = TableColocationId(targetRelationId);
+
 	/* finally set colocation group for the target relation */
 	UpdateRelationColocationGroup(targetRelationId, sourceColocationId);
 
+	/* if there is not any remaining table in the colocation group, delete it */
+	DeleteColocationGroupIfNoTablesBelong(targetColocationId);
+
 	heap_close(pgDistColocation, NoLock);
 }
 
@@ -277,16 +284,70 @@ ErrorIfShardPlacementsNotColocated(Oid leftRelationId, Oid rightRelationId)
 
 
 /*
- * ShardsIntervalsEqual checks if two shard intervals of hash distributed
+ * ShardsIntervalsEqual checks if two shard intervals of distributed
+ * tables are equal.
+ *
+ * Notes on the function:
+ * (i)   The function returns true if both shard intervals are the same.
+ * (ii)  The function returns false even if the shard intervals equal, but,
+ *       their distribution method are different.
+ * (iii) The function returns false for append and range partitioned tables
+ *       excluding (i) case.
+ * (iv)  For reference tables, all shards are equal (i.e., same replication factor
+ *       and shard min/max values). Thus, always return true for shards of reference
+ *       tables.
+ */
+static bool
+ShardsIntervalsEqual(ShardInterval *leftShardInterval, ShardInterval *rightShardInterval)
+{
+	char leftIntervalPartitionMethod = PartitionMethod(leftShardInterval->relationId);
+	char rightIntervalPartitionMethod = PartitionMethod(rightShardInterval->relationId);
+
+	/* if both shards are  the same, return true */
+	if (leftShardInterval->shardId == rightShardInterval->shardId)
+	{
+		return true;
+	}
+
+	/* if partition methods are not the same, shards cannot be considered as co-located */
+	leftIntervalPartitionMethod = PartitionMethod(leftShardInterval->relationId);
+	rightIntervalPartitionMethod = PartitionMethod(rightShardInterval->relationId);
+	if (leftIntervalPartitionMethod != rightIntervalPartitionMethod)
+	{
+		return false;
+	}
+
+	if (leftIntervalPartitionMethod == DISTRIBUTE_BY_HASH)
+	{
+		return HashPartitionedShardIntervalsEqual(leftShardInterval, rightShardInterval);
+	}
+	else if (leftIntervalPartitionMethod == DISTRIBUTE_BY_NONE)
+	{
+		/*
+		 * Reference tables has only a single shard and all reference tables
+		 * are always co-located with each other.
+		 */
+
+		return true;
+	}
+
+	/* append and range partitioned shard never co-located */
+	return false;
+}
+
+
+/*
+ * HashPartitionedShardIntervalsEqual checks if two shard intervals of hash distributed
  * tables are equal. Note that, this function doesn't work with non-hash
  * partitioned table's shards.
  *
  * We do min/max value check here to decide whether two shards are colocated,
- * instead we can simply use FindShardIntervalIndex function on both shards then
+ * instead we can simply use ShardIndex function on both shards then
  * but do index check, but we avoid it because this way it is more cheaper.
  */
 static bool
-ShardsIntervalsEqual(ShardInterval *leftShardInterval, ShardInterval *rightShardInterval)
+HashPartitionedShardIntervalsEqual(ShardInterval *leftShardInterval,
+								   ShardInterval *rightShardInterval)
 {
 	int32 leftShardMinValue = DatumGetInt32(leftShardInterval->minValue);
 	int32 leftShardMaxValue = DatumGetInt32(leftShardInterval->maxValue);
@@ -339,6 +400,208 @@ CompareShardPlacementsByNode(const void *leftElement, const void *rightElement)
 }
 
 
+/*
+ * ColocationId searches pg_dist_colocation for shard count, replication factor
+ * and distribution column type. If a matching entry is found, it returns the
+ * colocation id, otherwise it returns INVALID_COLOCATION_ID.
+ */
+uint32
+ColocationId(int shardCount, int replicationFactor, Oid distributionColumnType)
+{
+	uint32 colocationId = INVALID_COLOCATION_ID;
+	HeapTuple colocationTuple = NULL;
+	SysScanDesc scanDescriptor;
+	const int scanKeyCount = 3;
+	ScanKeyData scanKey[scanKeyCount];
+	bool indexOK = true;
+
+	Relation pgDistColocation = heap_open(DistColocationRelationId(), AccessShareLock);
+
+	/* set scan arguments */
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_colocation_shardcount,
+				BTEqualStrategyNumber, F_INT4EQ, UInt32GetDatum(shardCount));
+	ScanKeyInit(&scanKey[1], Anum_pg_dist_colocation_replicationfactor,
+				BTEqualStrategyNumber, F_INT4EQ, Int32GetDatum(replicationFactor));
+	ScanKeyInit(&scanKey[2], Anum_pg_dist_colocation_distributioncolumntype,
+				BTEqualStrategyNumber, F_OIDEQ, ObjectIdGetDatum(distributionColumnType));
+
+	scanDescriptor = systable_beginscan(pgDistColocation,
+										DistColocationConfigurationIndexId(),
+										indexOK, NULL, scanKeyCount, scanKey);
+
+	colocationTuple = systable_getnext(scanDescriptor);
+	if (HeapTupleIsValid(colocationTuple))
+	{
+		Form_pg_dist_colocation colocationForm =
+			(Form_pg_dist_colocation) GETSTRUCT(colocationTuple);
+
+		colocationId = colocationForm->colocationid;
+	}
+
+	systable_endscan(scanDescriptor);
+	heap_close(pgDistColocation, AccessShareLock);
+
+	return colocationId;
+}
+
+
+/*
+ * CreateColocationGroup creates a new colocation id and writes it into
+ * pg_dist_colocation with the given configuration. It also returns the created
+ * colocation id.
+ */
+uint32
+CreateColocationGroup(int shardCount, int replicationFactor, Oid distributionColumnType)
+{
+	uint32 colocationId = GetNextColocationId();
+	Relation pgDistColocation = NULL;
+	TupleDesc tupleDescriptor = NULL;
+	HeapTuple heapTuple = NULL;
+	Datum values[Natts_pg_dist_colocation];
+	bool isNulls[Natts_pg_dist_colocation];
+
+	/* form new colocation tuple */
+	memset(values, 0, sizeof(values));
+	memset(isNulls, false, sizeof(isNulls));
+
+	values[Anum_pg_dist_colocation_colocationid - 1] = UInt32GetDatum(colocationId);
+	values[Anum_pg_dist_colocation_shardcount - 1] = UInt32GetDatum(shardCount);
+	values[Anum_pg_dist_colocation_replicationfactor - 1] =
+		UInt32GetDatum(replicationFactor);
+	values[Anum_pg_dist_colocation_distributioncolumntype - 1] =
+		ObjectIdGetDatum(distributionColumnType);
+
+	/* open colocation relation and insert the new tuple */
+	pgDistColocation = heap_open(DistColocationRelationId(), RowExclusiveLock);
+
+	tupleDescriptor = RelationGetDescr(pgDistColocation);
+	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
+
+	simple_heap_insert(pgDistColocation, heapTuple);
+	CatalogUpdateIndexes(pgDistColocation, heapTuple);
+
+	/* increment the counter so that next command can see the row */
+	CommandCounterIncrement();
+	heap_close(pgDistColocation, RowExclusiveLock);
+
+	return colocationId;
+}
+
+
+/*
+ * GetNextColocationId allocates and returns a unique colocationId for the
+ * colocation group to be created. This allocation occurs both in shared memory
+ * and in write ahead logs; writing to logs avoids the risk of having
+ * colocationId collisions.
+ *
+ * Please note that the caller is still responsible for finalizing colocationId
+ * with the master node. Further note that this function relies on an internal
+ * sequence created in initdb to generate unique identifiers.
+ */
+uint32
+GetNextColocationId()
+{
+	text *sequenceName = cstring_to_text(COLOCATIONID_SEQUENCE_NAME);
+	Oid sequenceId = ResolveRelationId(sequenceName);
+	Datum sequenceIdDatum = ObjectIdGetDatum(sequenceId);
+	Oid savedUserId = InvalidOid;
+	int savedSecurityContext = 0;
+	Datum colocationIdDatum = 0;
+	uint32 colocationId = INVALID_COLOCATION_ID;
+
+	GetUserIdAndSecContext(&savedUserId, &savedSecurityContext);
+	SetUserIdAndSecContext(CitusExtensionOwner(), SECURITY_LOCAL_USERID_CHANGE);
+
+	/* generate new and unique colocation id from sequence */
+	colocationIdDatum = DirectFunctionCall1(nextval_oid, sequenceIdDatum);
+
+	SetUserIdAndSecContext(savedUserId, savedSecurityContext);
+
+	colocationId = DatumGetUInt32(colocationIdDatum);
+
+	return colocationId;
+}
+
+
+/*
+ * CheckReplicationModel checks if given relations are from the same
+ * replication model. Otherwise, it errors out.
+ */
+void
+CheckReplicationModel(Oid sourceRelationId, Oid targetRelationId)
+{
+	DistTableCacheEntry *sourceTableEntry = NULL;
+	DistTableCacheEntry *targetTableEntry = NULL;
+	char sourceReplicationModel = 0;
+	char targetReplicationModel = 0;
+
+	sourceTableEntry = DistributedTableCacheEntry(sourceRelationId);
+	sourceReplicationModel = sourceTableEntry->replicationModel;
+
+	targetTableEntry = DistributedTableCacheEntry(targetRelationId);
+	targetReplicationModel = targetTableEntry->replicationModel;
+
+	if (sourceReplicationModel != targetReplicationModel)
+	{
+		char *sourceRelationName = get_rel_name(sourceRelationId);
+		char *targetRelationName = get_rel_name(targetRelationId);
+
+		ereport(ERROR, (errmsg("cannot colocate tables %s and %s",
+							   sourceRelationName, targetRelationName),
+						errdetail("Replication models don't match for %s and %s.",
+								  sourceRelationName, targetRelationName)));
+	}
+}
+
+
+/*
+ * CheckDistributionColumnType checks if distribution column types of relations
+ * are same. Otherwise, it errors out.
+ */
+void
+CheckDistributionColumnType(Oid sourceRelationId, Oid targetRelationId)
+{
+	Var *sourceDistributionColumn = NULL;
+	Var *targetDistributionColumn = NULL;
+	Oid sourceDistributionColumnType = InvalidOid;
+	Oid targetDistributionColumnType = InvalidOid;
+
+	/* reference tables have NULL distribution column */
+	sourceDistributionColumn = PartitionKey(sourceRelationId);
+	if (sourceDistributionColumn == NULL)
+	{
+		sourceDistributionColumnType = InvalidOid;
+	}
+	else
+	{
+		sourceDistributionColumnType = sourceDistributionColumn->vartype;
+	}
+
+	/* reference tables have NULL distribution column */
+	targetDistributionColumn = PartitionKey(targetRelationId);
+	if (targetDistributionColumn == NULL)
+	{
+		targetDistributionColumnType = InvalidOid;
+	}
+	else
+	{
+		targetDistributionColumnType = targetDistributionColumn->vartype;
+	}
+
+	if (sourceDistributionColumnType != targetDistributionColumnType)
+	{
+		char *sourceRelationName = get_rel_name(sourceRelationId);
+		char *targetRelationName = get_rel_name(targetRelationId);
+
+		ereport(ERROR, (errmsg("cannot colocate tables %s and %s",
+							   sourceRelationName, targetRelationName),
+						errdetail("Distribution column types don't match for "
+								  "%s and %s.", sourceRelationName,
+								  targetRelationName)));
+	}
+}
+
+
 /*
  * UpdateRelationColocationGroup updates colocation group in pg_dist_partition
  * for the given relation.
@@ -350,6 +613,7 @@ UpdateRelationColocationGroup(Oid distributedRelationId, uint32 colocationId)
 	HeapTuple heapTuple = NULL;
 	TupleDesc tupleDescriptor = NULL;
 	SysScanDesc scanDescriptor = NULL;
+	bool shouldSyncMetadata = false;
 	bool indexOK = true;
 	int scanKeyCount = 1;
 	ScanKeyData scanKey[scanKeyCount];
@@ -375,7 +639,7 @@ UpdateRelationColocationGroup(Oid distributedRelationId, uint32 colocationId)
 							   distributedRelationName)));
 	}
 
-	memset(values, 0, sizeof(replace));
+	memset(values, 0, sizeof(values));
 	memset(isNull, false, sizeof(isNull));
 	memset(replace, false, sizeof(replace));
 
@@ -393,6 +657,15 @@ UpdateRelationColocationGroup(Oid distributedRelationId, uint32 colocationId)
 
 	systable_endscan(scanDescriptor);
 	heap_close(pgDistPartition, NoLock);
+
+	shouldSyncMetadata = ShouldSyncTableMetadata(distributedRelationId);
+	if (shouldSyncMetadata)
+	{
+		char *updateColocationIdCommand = ColocationIdUpdateCommand(distributedRelationId,
+																	colocationId);
+
+		SendCommandToWorkers(WORKERS_WITH_METADATA, updateColocationIdCommand);
+	}
 }
 
 
@@ -475,6 +748,30 @@ ColocatedTableList(Oid distributedTableId)
 	uint32 tableColocationId = TableColocationId(distributedTableId);
 	List *colocatedTableList = NIL;
 
+	/*
+	 * If distribution type of the table is not hash, the table is only co-located
+	 * with itself.
+	 */
+	if (tableColocationId == INVALID_COLOCATION_ID)
+	{
+		colocatedTableList = lappend_oid(colocatedTableList, distributedTableId);
+		return colocatedTableList;
+	}
+
+	colocatedTableList = ColocationGroupTableList(tableColocationId);
+
+	return colocatedTableList;
+}
+
+
+/*
+ * ColocationGroupTableList returns the list of tables in the given colocation
+ * group. If the colocation group is INVALID_COLOCATION_ID, it returns NIL.
+ */
+static List *
+ColocationGroupTableList(Oid colocationId)
+{
+	List *colocatedTableList = NIL;
 	Relation pgDistPartition = NULL;
 	TupleDesc tupleDescriptor = NULL;
 	SysScanDesc scanDescriptor = NULL;
@@ -487,14 +784,13 @@ ColocatedTableList(Oid distributedTableId)
 	 * If distribution type of the table is not hash, the table is only co-located
 	 * with itself.
 	 */
-	if (tableColocationId == INVALID_COLOCATION_ID)
+	if (colocationId == INVALID_COLOCATION_ID)
 	{
-		colocatedTableList = lappend_oid(colocatedTableList, distributedTableId);
-		return colocatedTableList;
+		return NIL;
 	}
 
 	ScanKeyInit(&scanKey[0], Anum_pg_dist_partition_colocationid,
-				BTEqualStrategyNumber, F_INT4EQ, ObjectIdGetDatum(tableColocationId));
+				BTEqualStrategyNumber, F_INT4EQ, ObjectIdGetDatum(colocationId));
 
 	pgDistPartition = heap_open(DistPartitionRelationId(), AccessShareLock);
 	tupleDescriptor = RelationGetDescr(pgDistPartition);
@@ -540,19 +836,24 @@ ColocatedShardIntervalList(ShardInterval *shardInterval)
 	char partitionMethod = cacheEntry->partitionMethod;
 
 	/*
-	 * If distribution type of the table is not hash, each shard of the table is only
-	 * co-located with itself.
+	 * If distribution type of the table is not hash or reference, each shard of
+	 * the shard is only co-located with itself.
 	 */
-	if (partitionMethod != DISTRIBUTE_BY_HASH)
+	if ((partitionMethod == DISTRIBUTE_BY_APPEND) ||
+		(partitionMethod == DISTRIBUTE_BY_RANGE))
 	{
-		colocatedShardList = lappend(colocatedShardList, shardInterval);
+		ShardInterval *copyShardInterval = CitusMakeNode(ShardInterval);
+		CopyShardInterval(shardInterval, copyShardInterval);
+
+		colocatedShardList = lappend(colocatedShardList, copyShardInterval);
+
 		return colocatedShardList;
 	}
 
-	shardIntervalIndex = FindShardIntervalIndex(shardInterval);
+	shardIntervalIndex = ShardIndex(shardInterval);
 	colocatedTableList = ColocatedTableList(distributedTableId);
 
-	/* FindShardIntervalIndex have to find index of given shard */
+	/* ShardIndex have to find index of given shard */
 	Assert(shardIntervalIndex >= 0);
 
 	foreach(colocatedTableCell, colocatedTableList)
@@ -561,6 +862,7 @@ ColocatedShardIntervalList(ShardInterval *shardInterval)
 		DistTableCacheEntry *colocatedTableCacheEntry =
 			DistributedTableCacheEntry(colocatedTableId);
 		ShardInterval *colocatedShardInterval = NULL;
+		ShardInterval *copyShardInterval = NULL;
 
 		/*
 		 * Since we iterate over co-located tables, shard count of each table should be
@@ -572,7 +874,10 @@ ColocatedShardIntervalList(ShardInterval *shardInterval)
 		colocatedShardInterval =
 			colocatedTableCacheEntry->sortedShardIntervalArray[shardIntervalIndex];
 
-		colocatedShardList = lappend(colocatedShardList, colocatedShardInterval);
+		copyShardInterval = CitusMakeNode(ShardInterval);
+		CopyShardInterval(colocatedShardInterval, copyShardInterval);
+
+		colocatedShardList = lappend(colocatedShardList, copyShardInterval);
 	}
 
 	Assert(list_length(colocatedTableList) == list_length(colocatedShardList));
@@ -633,3 +938,61 @@ ColocatedShardIdInRelation(Oid relationId, int shardIndex)
 
 	return tableCacheEntry->sortedShardIntervalArray[shardIndex]->shardId;
 }
+
+
+/*
+ * DeleteColocationGroupIfNoTablesBelong function deletes given co-location group if there
+ * is no relation in that co-location group. A co-location group may become empty after
+ * mark_tables_colocated or upgrade_reference_table UDF calls. In that case we need to
+ * remove empty co-location group to prevent orphaned co-location groups.
+ */
+void
+DeleteColocationGroupIfNoTablesBelong(uint32 colocationId)
+{
+	if (colocationId != INVALID_COLOCATION_ID)
+	{
+		List *colocatedTableList = ColocationGroupTableList(colocationId);
+		int colocatedTableCount = list_length(colocatedTableList);
+
+		if (colocatedTableCount == 0)
+		{
+			DeleteColocationGroup(colocationId);
+		}
+	}
+}
+
+
+/*
+ * DeleteColocationGroup deletes the colocation group from pg_dist_colocation.
+ */
+static void
+DeleteColocationGroup(uint32 colocationId)
+{
+	Relation pgDistColocation = NULL;
+	SysScanDesc scanDescriptor = NULL;
+	int scanKeyCount = 1;
+	ScanKeyData scanKey[scanKeyCount];
+	bool indexOK = false;
+	HeapTuple heapTuple = NULL;
+
+	pgDistColocation = heap_open(DistColocationRelationId(), RowExclusiveLock);
+
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_colocation_colocationid,
+				BTEqualStrategyNumber, F_INT4EQ, UInt32GetDatum(colocationId));
+
+	scanDescriptor = systable_beginscan(pgDistColocation, InvalidOid, indexOK,
+										NULL, scanKeyCount, scanKey);
+
+	/* if a record is found, delete it */
+	heapTuple = systable_getnext(scanDescriptor);
+	if (HeapTupleIsValid(heapTuple))
+	{
+		simple_heap_delete(pgDistColocation, &(heapTuple->t_self));
+
+		CitusInvalidateRelcacheByRelid(DistColocationRelationId());
+		CommandCounterIncrement();
+	}
+
+	systable_endscan(scanDescriptor);
+	heap_close(pgDistColocation, RowExclusiveLock);
+}
