@@ -67,8 +67,12 @@ task_tracker_assign_task(PG_FUNCTION_ARGS)
 	char *taskCallString = text_to_cstring(taskCallStringText);
 	uint32 taskCallStringLength = strlen(taskCallString);
 
+	bool taskTrackerRunning = false;
+
+	CheckCitusVersion(ERROR);
+
 	/* check that we have a running task tracker on this host */
-	bool taskTrackerRunning = TaskTrackerRunning();
+	taskTrackerRunning = TaskTrackerRunning();
 	if (!taskTrackerRunning)
 	{
 		ereport(ERROR, (errcode(ERRCODE_CANNOT_CONNECT_NOW),
@@ -129,7 +133,12 @@ task_tracker_task_status(PG_FUNCTION_ARGS)
 	WorkerTask *workerTask = NULL;
 	uint32 taskStatus = 0;
 
-	bool taskTrackerRunning = TaskTrackerRunning();
+	bool taskTrackerRunning = false;
+
+	CheckCitusVersion(ERROR);
+
+	taskTrackerRunning = TaskTrackerRunning();
+
 	if (taskTrackerRunning)
 	{
 		LWLockAcquire(&WorkerTasksSharedState->taskHashLock, LW_SHARED);
@@ -170,6 +179,8 @@ task_tracker_cleanup_job(PG_FUNCTION_ARGS)
 	StringInfo jobDirectoryName = NULL;
 	StringInfo jobSchemaName = NULL;
 
+	CheckCitusVersion(ERROR);
+
 	/*
 	 * We first clean up any open connections, and remove tasks belonging to
 	 * this job from the shared hash.
@@ -261,12 +272,8 @@ CreateJobSchema(StringInfo schemaName)
 
 	Oid savedUserId = InvalidOid;
 	int savedSecurityContext = 0;
-
-	/* build a CREATE SCHEMA statement */
-	CreateSchemaStmt *createSchemaStmt = makeNode(CreateSchemaStmt);
-	createSchemaStmt->schemaname = schemaName->data;
-	createSchemaStmt->authrole = NULL;
-	createSchemaStmt->schemaElts = NIL;
+	CreateSchemaStmt *createSchemaStmt = NULL;
+	RoleSpec currentUserRole = { 0 };
 
 	/* allow schema names that start with pg_ */
 	oldAllowSystemTableMods = allowSystemTableMods;
@@ -276,7 +283,18 @@ CreateJobSchema(StringInfo schemaName)
 	GetUserIdAndSecContext(&savedUserId, &savedSecurityContext);
 	SetUserIdAndSecContext(CitusExtensionOwner(), SECURITY_LOCAL_USERID_CHANGE);
 
-	/* actually create schema, and make it visible */
+	/* build a CREATE SCHEMA statement */
+	currentUserRole.type = T_RoleSpec;
+	currentUserRole.roletype = ROLESPEC_CSTRING;
+	currentUserRole.rolename = GetUserNameFromId(savedUserId, false);
+	currentUserRole.location = -1;
+
+	createSchemaStmt = makeNode(CreateSchemaStmt);
+	createSchemaStmt->schemaname = schemaName->data;
+	createSchemaStmt->authrole = (Node *) &currentUserRole;
+	createSchemaStmt->schemaElts = NIL;
+
+	/* actually create schema with the current user as owner */
 	CreateSchemaCommand(createSchemaStmt, queryString);
 	CommandCounterIncrement();
 
