@@ -23,6 +23,7 @@
 #include "catalog/pg_namespace.h"
 #include "commands/copy.h"
 #include "commands/tablecmds.h"
+#include "distributed/metadata_cache.h"
 #include "distributed/worker_protocol.h"
 #include "executor/spi.h"
 #include "nodes/makefuncs.h"
@@ -74,10 +75,15 @@ worker_merge_files_into_table(PG_FUNCTION_ARGS)
 	bool schemaExists = false;
 	List *columnNameList = NIL;
 	List *columnTypeList = NIL;
+	Oid savedUserId = InvalidOid;
+	int savedSecurityContext = 0;
 
 	/* we should have the same number of column names and types */
 	int32 columnNameCount = ArrayObjectCount(columnNameObject);
 	int32 columnTypeCount = ArrayObjectCount(columnTypeObject);
+
+	CheckCitusVersion(ERROR);
+
 	if (columnNameCount != columnTypeCount)
 	{
 		ereport(ERROR, (errmsg("column name array size: %d and type array size: %d"
@@ -101,8 +107,14 @@ worker_merge_files_into_table(PG_FUNCTION_ARGS)
 
 	CreateTaskTable(jobSchemaName, taskTableName, columnNameList, columnTypeList);
 
+	/* need superuser to copy from files */
+	GetUserIdAndSecContext(&savedUserId, &savedSecurityContext);
+	SetUserIdAndSecContext(CitusExtensionOwner(), SECURITY_LOCAL_USERID_CHANGE);
+
 	CopyTaskFilesFromDirectory(jobSchemaName, taskTableName, taskDirectoryName);
 
+	SetUserIdAndSecContext(savedUserId, savedSecurityContext);
+
 	PG_RETURN_VOID();
 }
 
@@ -143,6 +155,8 @@ worker_merge_files_and_run_query(PG_FUNCTION_ARGS)
 	int createIntermediateTableResult = 0;
 	int finished = 0;
 
+	CheckCitusVersion(ERROR);
+
 	/*
 	 * If the schema for the job isn't already created by the task tracker
 	 * protocol, we fall to using the default 'public' schema.
@@ -217,6 +231,8 @@ worker_cleanup_job_schema_cache(PG_FUNCTION_ARGS)
 	int scanKeyCount = 0;
 	HeapTuple heapTuple = NULL;
 
+	CheckCitusVersion(ERROR);
+
 	pgNamespace = heap_open(NamespaceRelationId, AccessExclusiveLock);
 	scanDescriptor = heap_beginscan_catalog(pgNamespace, scanKeyCount, scanKey);
 
