@@ -18,7 +18,9 @@
 #include "c.h"
 #include "fmgr.h"
 
+#include "distributed/shardinterval_utils.h"
 #include "nodes/pg_list.h"
+#include "distributed/master_metadata_utility.h"
 
 
 /*
@@ -66,10 +68,10 @@
 #define SHARD_CSTORE_TABLE_SIZE_QUERY "SELECT cstore_table_size(%s)"
 #define DROP_REGULAR_TABLE_COMMAND "DROP TABLE IF EXISTS %s CASCADE"
 #define DROP_FOREIGN_TABLE_COMMAND "DROP FOREIGN TABLE IF EXISTS %s CASCADE"
-#define CREATE_SCHEMA_COMMAND "CREATE SCHEMA IF NOT EXISTS %s"
+#define CREATE_SCHEMA_COMMAND "CREATE SCHEMA IF NOT EXISTS %s AUTHORIZATION %s"
 #define CREATE_EMPTY_SHARD_QUERY "SELECT master_create_empty_shard('%s')"
 #define FINALIZED_SHARD_PLACEMENTS_QUERY \
-	"SELECT nodename, nodeport FROM pg_dist_shard_placement WHERE shardstate = 1 AND shardid = %ld"
+	"SELECT placementid, nodename, nodeport FROM pg_dist_shard_placement WHERE shardstate = 1 AND shardid = %ld"
 #define UPDATE_SHARD_STATISTICS_QUERY \
 	"SELECT master_update_shard_statistics(%ld)"
 #define PARTITION_METHOD_QUERY "SELECT part_method FROM master_get_table_metadata('%s');"
@@ -91,11 +93,16 @@ extern int ShardMaxSize;
 extern int ShardPlacementPolicy;
 
 
+extern bool IsCoordinator(void);
+
 /* Function declarations local to the distributed module */
 extern bool CStoreTable(Oid relationId);
 extern uint64 GetNextShardId(void);
+extern uint64 GetNextPlacementId(void);
 extern Oid ResolveRelationId(text *relationName);
-extern List * GetTableDDLEvents(Oid relationId);
+extern List * GetTableDDLEvents(Oid relationId, bool forShardCreation);
+extern List * GetTableCreationCommands(Oid relationId, bool forShardCreation);
+extern List * GetTableIndexAndConstraintCommands(Oid relationId);
 extern List * GetTableForeignConstraintCommands(Oid relationId);
 extern char ShardStorageType(Oid relationId);
 extern void CheckDistributedTable(Oid relationId);
@@ -106,19 +113,21 @@ extern uint64 UpdateShardStatistics(int64 shardId);
 extern void CreateShardsWithRoundRobinPolicy(Oid distributedTableId, int32 shardCount,
 											 int32 replicationFactor);
 extern void CreateColocatedShards(Oid targetRelationId, Oid sourceRelationId);
+extern void CreateReferenceTableShard(Oid distributedTableId);
 extern bool WorkerCreateShard(Oid relationId, char *nodeName, uint32 nodePort,
 							  int shardIndex, uint64 shardId, char *newShardOwner,
 							  List *ddlCommandList, List *foreignConstraintCommadList);
 extern Oid ForeignConstraintGetReferencedTableId(char *queryString);
 extern void CheckHashPartitionedTable(Oid distributedTableId);
+extern void CheckTableSchemaNameForDrop(Oid relationId, char **schemaName,
+										char **tableName);
+extern text * IntegerToText(int32 value);
 
 /* Function declarations for generating metadata for shard and placement creation */
 extern Datum master_get_table_metadata(PG_FUNCTION_ARGS);
 extern Datum master_get_table_ddl_events(PG_FUNCTION_ARGS);
 extern Datum master_get_new_shardid(PG_FUNCTION_ARGS);
 extern Datum master_get_new_placementid(PG_FUNCTION_ARGS);
-extern Datum master_get_local_first_candidate_nodes(PG_FUNCTION_ARGS);
-extern Datum master_get_round_robin_candidate_nodes(PG_FUNCTION_ARGS);
 extern Datum master_get_active_worker_nodes(PG_FUNCTION_ARGS);
 
 /* Function declarations to help with data staging and deletion */
@@ -132,9 +141,17 @@ extern Datum master_drop_all_shards(PG_FUNCTION_ARGS);
 
 /* function declarations for shard creation functionality */
 extern Datum master_create_worker_shards(PG_FUNCTION_ARGS);
+extern Datum isolate_tenant_to_new_shard(PG_FUNCTION_ARGS);
 
 /* function declarations for shard repair functionality */
 extern Datum master_copy_shard_placement(PG_FUNCTION_ARGS);
 
+/* function declarations for shard copy functinality */
+extern List * CopyShardCommandList(ShardInterval *shardInterval, char *sourceNodeName,
+								   int32 sourceNodePort);
+extern List * CopyShardForeignConstraintCommandList(ShardInterval *shardInterval);
+extern ShardPlacement * SearchShardPlacementInList(List *shardPlacementList,
+												   char *nodeName, uint32 nodePort,
+												   bool missingOk);
 
 #endif   /* MASTER_PROTOCOL_H */
