@@ -26,19 +26,26 @@ sub Usage()
     print "  pg_regress_multi [MULTI OPTIONS] -- [PG REGRESS OPTS]\n";
     print "\n";
     print "Multi Options:\n";
-    print "  --isolationtester   Run isolationtester tests instead of plain tests\n";
-    print "  --bindir            Path to postgres binary directory\n";
-    print "  --libdir            Path to postgres library directory\n";
-    print "  --postgres-builddir Path to postgres build directory\n";
-    print "  --postgres-srcdir Path to postgres build directory\n";
-    print "  --pgxsdir           Path to the PGXS directory\n";
-    print "  --load-extension    Extensions to install in all nodes\n";
-    print "  --server-option     Config option to pass to the server\n";
+    print "  --isolationtester   	Run isolationtester tests instead of plain tests\n";
+    print "  --vanillatest       	Run postgres tests with citus loaded as shared preload library\n";
+    print "  --bindir            	Path to postgres binary directory\n";
+    print "  --libdir            	Path to postgres library directory\n";
+    print "  --postgres-builddir 	Path to postgres build directory\n";
+    print "  --postgres-srcdir   	Path to postgres build directory\n";
+    print "  --pgxsdir           	Path to the PGXS directory\n";
+    print "  --load-extension    	Extensions to install in all nodes\n";
+    print "  --server-option     	Config option to pass to the server\n";
+    print "  --valgrind          	Run server via valgrind\n";
+    print "  --valgrind-path     	Path to the valgrind executable\n";
+    print "  --valgrind-log-file	Path to the write valgrind logs\n";
+    print "  --pg_ctl-timeout    	Timeout for pg_ctl\n";
+    print "  --connection-timeout	Timeout for connecting to worker nodes\n";
     exit 1;
 }
 
 # Option parsing
 my $isolationtester = 0;
+my $vanillatest = 0;
 my $bindir = "";
 my $libdir = undef;
 my $pgxsdir = "";
@@ -52,11 +59,17 @@ my %fdws = ();
 my %fdwServers = ();
 my %functions = ();
 my %operators = ();
+my $valgrind = 0;
+my $valgrindPath = "valgrind";
+my $valgrindLogFile = "valgrind_test_log.txt";
+my $pgCtlTimeout = undef;
+my $connectionTimeout = 5000;
 
 my $serversAreShutdown = "TRUE";
 
 GetOptions(
     'isolationtester' => \$isolationtester,
+    'vanillatest' => \$vanillatest,
     'bindir=s' => \$bindir,
     'libdir=s' => \$libdir,
     'pgxsdir=s' => \$pgxsdir,
@@ -65,6 +78,11 @@ GetOptions(
     'majorversion=s' => \$majorversion,
     'load-extension=s' => \@extensions,
     'server-option=s' => \@userPgOptions,
+    'valgrind' => \$valgrind,
+    'valgrind-path=s' => \$valgrindPath,
+    'valgrind-log-file=s' => \$valgrindLogFile,
+    'pg_ctl-timeout=s' => \$pgCtlTimeout,
+    'connection-timeout=s' => \$connectionTimeout,
     'help' => sub { Usage() });
 
 # Update environment to include [DY]LD_LIBRARY_PATH/LIBDIR/etc -
@@ -82,6 +100,21 @@ if (defined $libdir)
     $ENV{PATH} = "$libdir:".($ENV{PATH} || '');
 }
 
+# Put $bindir to the end of PATH. We want to prefer system binaries by
+# default (as e.g. new libpq and old psql can cause issues), but still
+# want to find binaries if they're not in PATH.
+if (defined $bindir)
+{
+    $ENV{PATH} = ($ENV{PATH} || '').":$bindir";
+}
+
+
+# Most people are used to unified diffs these days, rather than the
+# context diffs pg_regress defaults to.  Change default to avoid
+# everyone having to (re-)learn how to change that setting.  Also add
+# a bit more context to make it easier to locate failed test sections.
+$ENV{PG_REGRESS_DIFF_OPTS} = '-dU10';
+
 my $plainRegress = "$pgxsdir/src/test/regress/pg_regress";
 my $isolationRegress = "${postgresBuilddir}/src/test/isolation/pg_isolation_regress";
 if ($isolationtester && ! -f "$isolationRegress")
@@ -101,6 +134,78 @@ directory is present locally
 MESSAGE
 }
 
+my $vanillaRegress = "${postgresBuilddir}/src/test/regress/pg_regress";
+if ($vanillatest && ! -f "$vanillaRegress")
+{
+    die <<"MESSAGE";
+
+pg_regress (for vanilla tests) not found at $vanillaRegress.
+
+Vanilla tests can only be run when source (detected as ${postgresSrcdir})
+and build (detected as ${postgresBuilddir}) directory corresponding to $bindir
+are present.
+MESSAGE
+}
+
+# If pgCtlTimeout is defined, we will set related environment variable.
+# This is generally used with valgrind because valgrind starts slow and we
+# need to increase timeout.
+if (defined $pgCtlTimeout)
+{
+    $ENV{PGCTLTIMEOUT} = "$pgCtlTimeout";
+}
+
+# We don't want valgrind to run pg_ctl itself, as that'd trigger a lot
+# of spurious OS failures, e.g. in bash. So instead we have to replace
+# the postgres binary with a wrapper that exec's valgrind, which in
+# turn then executes postgres.  That's unfortunately at the moment the
+# only reliable way to do this.
+sub replace_postgres
+{
+    if (-e "$bindir/postgres.orig")
+    {
+	print "wrapper exists\n";
+    }
+    else
+    {
+	print "moving $bindir/postgres to $bindir/postgres.orig\n";
+	rename "$bindir/postgres", "$bindir/postgres.orig"
+	    or die "Could not move postgres out of the way";
+    }
+
+    sysopen my $fh, "$bindir/postgres", O_CREAT|O_TRUNC|O_RDWR, 0700
+	or die "Could not create postgres wrapper at $bindir/postgres";
+    print $fh <<"END";
+#!/bin/bash
+exec $valgrindPath \\
+    --quiet \\
+    --suppressions=${postgresSrcdir}/src/tools/valgrind.supp \\
+    --trace-children=yes --track-origins=yes --read-var-info=no \\
+    --leak-check=no \\
+    --error-markers=VALGRINDERROR-BEGIN,VALGRINDERROR-END \\
+    --log-file=$valgrindLogFile \\
+    $bindir/postgres.orig \\
+    "\$@"
+END
+    close $fh;
+}
+
+# revert changes replace_postgres() performed
+sub revert_replace_postgres
+{
+    if (-e "$bindir/postgres.orig")
+    {
+	print "wrapper exists, removing\n";
+	print "moving $bindir/postgres.orig to $bindir/postgres\n";
+	rename "$bindir/postgres.orig", "$bindir/postgres"
+	    or die "Could not move postgres back";
+    }
+}
+
+# always want to call initdb under normal postgres, so revert from a
+# partial run, even if we're now not using valgrind.
+revert_replace_postgres();
+
 # Set some default configuration options
 my $masterPort = 57636;
 my $workerCount = 2;
@@ -121,7 +226,6 @@ push(@pgOptions, '-c', "listen_addresses='${host}'");
 push(@pgOptions, '-c', "unix_socket_directories=");
 push(@pgOptions, '-c', "fsync=off");
 push(@pgOptions, '-c', "shared_preload_libraries=citus");
-push(@pgOptions, '-c', "max_prepared_transactions=100");
 
 # Citus options set for the tests
 push(@pgOptions, '-c', "citus.shard_max_size=300kB");
@@ -129,6 +233,8 @@ push(@pgOptions, '-c', "citus.max_running_tasks_per_node=4");
 push(@pgOptions, '-c', "citus.expire_cached_shards=on");
 push(@pgOptions, '-c', "citus.task_tracker_delay=10ms");
 push(@pgOptions, '-c', "citus.remote_task_check_interval=1ms");
+push(@pgOptions, '-c', "citus.shard_replication_factor=2");
+push(@pgOptions, '-c', "citus.node_connection_timeout=${connectionTimeout}");
 
 # Add externally added options last, so they overwrite the default ones above
 for my $option (@userPgOptions)
@@ -137,7 +243,7 @@ for my $option (@userPgOptions)
 }
 
 #define data types as a name->definition
-%dataTypes = ('dummy_type', '(i integer)', 
+%dataTypes = ('dummy_type', '(i integer)',
                'order_side', ' ENUM (\'buy\', \'sell\')',
                'test_composite_type', '(i integer, i2 integer)',
                'bug_status', ' ENUM (\'new\', \'open\', \'closed\')');
@@ -171,8 +277,9 @@ system("mkdir", ('-p', "tmp_check/tmp-bin")) == 0
 sysopen my $fh, "tmp_check/tmp-bin/psql", O_CREAT|O_TRUNC|O_RDWR, 0700
 	or die "Could not create psql wrapper";
 print $fh "#!/bin/bash\n";
-print $fh "exec $bindir/psql ";
+print $fh "exec psql ";
 print $fh "--variable=master_port=$masterPort ";
+print $fh "--variable=default_user=$user ";
 print $fh "--variable=SHOW_CONTEXT=always ";
 for my $workeroff (0 .. $#workerPorts)
 {
@@ -231,6 +338,18 @@ END
     {
         ShutdownServers();
     }
+
+    # At the end of a run, replace redirected binary with original again
+    if ($valgrind)
+    {
+	revert_replace_postgres();
+    }
+}
+
+# want to use valgrind, replace binary before starting server
+if ($valgrind)
+{
+    replace_postgres();
 }
 
 # Signal that servers should be shutdown
@@ -259,14 +378,14 @@ for my $port (@workerPorts)
 ###
 for my $port (@workerPorts)
 {
-    system("$bindir/psql", '-X',
+    system("psql", '-X',
            ('-h', $host, '-p', $port, '-U', $user, "postgres",
             '-c', "CREATE DATABASE regression;")) == 0
         or die "Could not create regression database on worker";
 
     for my $extension (@extensions)
     {
-        system("$bindir/psql", '-X',
+        system("psql", '-X',
                ('-h', $host, '-p', $port, '-U', $user, "regression",
                 '-c', "CREATE EXTENSION IF NOT EXISTS \"$extension\";")) == 0
             or die "Could not create extension on worker";
@@ -274,7 +393,7 @@ for my $port (@workerPorts)
 
     foreach my $dataType (keys %dataTypes)
     {
-        system("$bindir/psql", '-X',
+        system("psql", '-X',
                 ('-h', $host, '-p', $port, '-U', $user, "regression",
                  '-c', "CREATE TYPE $dataType AS $dataTypes{$dataType};")) == 0
             or die "Could not create TYPE $dataType on worker";
@@ -282,7 +401,7 @@ for my $port (@workerPorts)
 
     foreach my $function (keys %functions)
     {
-        system("$bindir/psql", '-X',
+        system("psql", '-X',
                 ('-h', $host, '-p', $port, '-U', $user, "regression",
                  '-c', "CREATE FUNCTION $function RETURNS $functions{$function};")) == 0
             or die "Could not create FUNCTION $function on worker";
@@ -290,7 +409,7 @@ for my $port (@workerPorts)
 
     foreach my $operator (keys %operators)
     {
-        system("$bindir/psql", '-X',
+        system("psql", '-X',
                 ('-h', $host, '-p', $port, '-U', $user, "regression",
                  '-c', "CREATE OPERATOR $operator $operators{$operator};")) == 0
             or die "Could not create OPERATOR $operator on worker";
@@ -298,7 +417,7 @@ for my $port (@workerPorts)
 
     foreach my $fdw (keys %fdws)
     {
-        system("$bindir/psql", '-X',
+        system("psql", '-X',
                 ('-h', $host, '-p', $port, '-U', $user, "regression",
                  '-c', "CREATE FOREIGN DATA WRAPPER $fdw HANDLER $fdws{$fdw};")) == 0
             or die "Could not create foreign data wrapper $fdw on worker";
@@ -306,7 +425,7 @@ for my $port (@workerPorts)
 
     foreach my $fdwServer (keys %fdwServers)
     {
-        system("$bindir/psql", '-X',
+        system("psql", '-X',
                 ('-h', $host, '-p', $port, '-U', $user, "regression",
                  '-c', "CREATE SERVER $fdwServer FOREIGN DATA WRAPPER $fdwServers{$fdwServer};")) == 0
             or die "Could not create server $fdwServer on worker";
@@ -341,17 +460,27 @@ push(@arguments, @ARGV);
 my $startTime = time();
 
 # Finally run the tests
-if (!$isolationtester)
+if ($vanillatest)
 {
-    system("$plainRegress", @arguments) == 0
-	or die "Could not run regression tests";
+    push(@arguments, "--schedule=${postgresSrcdir}/parallel_schedule");
+    $ENV{PGHOST} = $host;
+    $ENV{PGPORT} = $masterPort;
+    $ENV{PGUSER} = $user;
+
+    system("make -C $postgresBuilddir/src/test/regress installcheck-parallel") == 0
+    or die "Could not run vanilla tests";
 }
-else
+elsif ($isolationtester)
 {
     push(@arguments, "--dbname=regression");
     system("$isolationRegress", @arguments) == 0
 	or die "Could not run isolation tests";
 }
+else
+{
+    system("$plainRegress", @arguments) == 0
+	or die "Could not run regression tests";
+}
 
 my $endTime = time();
 
