@@ -8,7 +8,6 @@
 
 
 ALTER SEQUENCE pg_catalog.pg_dist_shardid_seq RESTART 790000;
-ALTER SEQUENCE pg_catalog.pg_dist_jobid_seq RESTART 790000;
 
 
 PREPARE prepared_test_1 AS
@@ -149,8 +148,7 @@ EXECUTE prepared_test_2;
 
 -- execute prepared statements with different parameters
 EXECUTE prepared_test_6(155);
--- FIXME: temporarily disabled
--- EXECUTE prepared_test_6(1555);
+EXECUTE prepared_test_6(1555);
 
 -- test router executor with parameterized non-partition columns
 
@@ -194,12 +192,12 @@ EXECUTE prepared_select(4, 40);
 EXECUTE prepared_select(5, 50);
 EXECUTE prepared_select(6, 60);
 
--- test that we don't crash on failing parameterized insert on the partition column
+-- Test that parameterized partition column for an insert is supported
 
 PREPARE prepared_partition_column_insert(bigint) AS
 INSERT INTO router_executor_table VALUES ($1, 'arsenous', '(1,10)');
 
--- we error out on the 6th execution
+-- execute 6 times to trigger prepared statement usage
 EXECUTE prepared_partition_column_insert(1);
 EXECUTE prepared_partition_column_insert(2);
 EXECUTE prepared_partition_column_insert(3);
@@ -282,10 +280,7 @@ EXECUTE prepared_router_partition_column_select(2);
 EXECUTE prepared_router_partition_column_select(3);
 EXECUTE prepared_router_partition_column_select(4);
 EXECUTE prepared_router_partition_column_select(5);
-
--- FIXME: 6th execution is failing. We don't want to run the failing test
--- because of changing output. After implementing this feature, uncomment this.
--- EXECUTE prepared_router_partition_column_select(6);
+EXECUTE prepared_router_partition_column_select(6);
 
 PREPARE prepared_router_non_partition_column_select(int) AS
 	SELECT
@@ -325,10 +320,7 @@ EXECUTE prepared_real_time_non_partition_column_select(20);
 EXECUTE prepared_real_time_non_partition_column_select(30);
 EXECUTE prepared_real_time_non_partition_column_select(40);
 EXECUTE prepared_real_time_non_partition_column_select(50);
-
--- FIXME: 6th execution is failing. We don't want to run the failing test
--- because of changing output. After implementing this feature, uncomment this.
--- EXECUTE prepared_real_time_non_partition_column_select(60);
+EXECUTE prepared_real_time_non_partition_column_select(60);
 
 PREPARE prepared_real_time_partition_column_select(int) AS
 	SELECT
@@ -348,10 +340,7 @@ EXECUTE prepared_real_time_partition_column_select(2);
 EXECUTE prepared_real_time_partition_column_select(3);
 EXECUTE prepared_real_time_partition_column_select(4);
 EXECUTE prepared_real_time_partition_column_select(5);
-
--- FIXME: 6th execution is failing. We don't want to run the failing test
--- because of changing output. After implementing this feature, uncomment this.
--- EXECUTE prepared_real_time_partition_column_select(6);
+EXECUTE prepared_real_time_partition_column_select(6);
 
 -- check task-tracker executor
 SET citus.task_executor_type TO 'task-tracker';
@@ -373,10 +362,7 @@ EXECUTE prepared_task_tracker_non_partition_column_select(20);
 EXECUTE prepared_task_tracker_non_partition_column_select(30);
 EXECUTE prepared_task_tracker_non_partition_column_select(40);
 EXECUTE prepared_task_tracker_non_partition_column_select(50);
-
--- FIXME: 6th execution is failing. We don't want to run the failing test
--- because of changing output. After implementing this feature, uncomment this.
--- EXECUTE prepared_task_tracker_non_partition_column_select(60);
+EXECUTE prepared_task_tracker_non_partition_column_select(60);
 
 PREPARE prepared_task_tracker_partition_column_select(int) AS
 	SELECT
@@ -396,10 +382,7 @@ EXECUTE prepared_task_tracker_partition_column_select(2);
 EXECUTE prepared_task_tracker_partition_column_select(3);
 EXECUTE prepared_task_tracker_partition_column_select(4);
 EXECUTE prepared_task_tracker_partition_column_select(5);
-
--- FIXME: 6th execution is failing. We don't want to run the failing test
--- because of changing output. After implementing this feature, uncomment this.
--- EXECUTE prepared_task_tracker_partition_column_select(6);
+EXECUTE prepared_task_tracker_partition_column_select(6);
 
 SET citus.task_executor_type TO 'real-time';
 
@@ -413,8 +396,7 @@ EXECUTE prepared_partition_parameter_update(2, 21);
 EXECUTE prepared_partition_parameter_update(3, 31);
 EXECUTE prepared_partition_parameter_update(4, 41);
 EXECUTE prepared_partition_parameter_update(5, 51);
--- This fails with an unexpected error message
-EXECUTE prepared_partition_parameter_update(5, 52);
+EXECUTE prepared_partition_parameter_update(6, 61);
 
 PREPARE prepared_non_partition_parameter_update(int, int) AS
 	UPDATE prepare_table SET value = $2 WHERE key = 0 AND value = $1;
@@ -439,8 +421,7 @@ EXECUTE prepared_partition_parameter_delete(2, 21);
 EXECUTE prepared_partition_parameter_delete(3, 31);
 EXECUTE prepared_partition_parameter_delete(4, 41);
 EXECUTE prepared_partition_parameter_delete(5, 51);
--- This fails with an unexpected error message
-EXECUTE prepared_partition_parameter_delete(0, 10);
+EXECUTE prepared_partition_parameter_delete(6, 61);
 
 PREPARE prepared_non_partition_parameter_delete(int) AS
 	DELETE FROM prepare_table WHERE key = 0 AND value = $1;
@@ -456,6 +437,44 @@ EXECUTE prepared_non_partition_parameter_delete(62);
 -- check after deletes
 SELECT * FROM prepare_table ORDER BY key, value;
 
+-- Testing parameters + function evaluation
+CREATE TABLE prepare_func_table (
+    key text,
+    value1 int,
+    value2 text,
+    value3 timestamptz DEFAULT now()
+);
+SELECT create_distributed_table('prepare_func_table', 'key');
+
+-- test function evaluation with parameters in an expression
+PREPARE prepared_function_evaluation_insert(int) AS
+	INSERT INTO prepare_func_table (key, value1) VALUES ($1+1, 0*random());
+
+-- execute 6 times to trigger prepared statement usage
+EXECUTE prepared_function_evaluation_insert(1);
+EXECUTE prepared_function_evaluation_insert(2);
+EXECUTE prepared_function_evaluation_insert(3);
+EXECUTE prepared_function_evaluation_insert(4);
+EXECUTE prepared_function_evaluation_insert(5);
+EXECUTE prepared_function_evaluation_insert(6);
+
+SELECT key, value1 FROM prepare_func_table ORDER BY key;
+TRUNCATE prepare_func_table;
+
+-- make it a bit harder: parameter wrapped in a function call
+PREPARE wrapped_parameter_evaluation(text,text[]) AS
+	INSERT INTO prepare_func_table (key,value2) VALUES ($1,array_to_string($2,''));
+
+EXECUTE wrapped_parameter_evaluation('key', ARRAY['value']);
+EXECUTE wrapped_parameter_evaluation('key', ARRAY['value']);
+EXECUTE wrapped_parameter_evaluation('key', ARRAY['value']);
+EXECUTE wrapped_parameter_evaluation('key', ARRAY['value']);
+EXECUTE wrapped_parameter_evaluation('key', ARRAY['value']);
+EXECUTE wrapped_parameter_evaluation('key', ARRAY['value']);
+
+SELECT key, value2 FROM prepare_func_table;
+
+DROP TABLE prepare_func_table;
 
 -- verify placement state updates invalidate shard state
 --
@@ -509,3 +528,5 @@ EXECUTE countsome; -- no replanning
 
 -- clean-up prepared statements
 DEALLOCATE ALL;
+
+DROP TABLE prepare_table;
