@@ -1,6 +1,5 @@
 
 ALTER SEQUENCE pg_catalog.pg_dist_shardid_seq RESTART 840000;
-ALTER SEQUENCE pg_catalog.pg_dist_jobid_seq RESTART 840000;
 
 
 -- ===================================================================
@@ -14,11 +13,27 @@ CREATE TABLE articles_hash (
 	word_count integer
 );
 
+CREATE TABLE articles_range (
+	id bigint NOT NULL,
+	author_id bigint NOT NULL,
+	title varchar(20) NOT NULL,
+	word_count integer
+);
+
+CREATE TABLE articles_append (
+	id bigint NOT NULL,
+	author_id bigint NOT NULL,
+	title varchar(20) NOT NULL,
+	word_count integer
+);
+
 -- Check for the existence of line 'DEBUG:  Creating router plan'
 -- to determine if router planner is used.
 
 -- this table is used in a CTE test
-CREATE TABLE authors_hash ( name text, id bigint );
+CREATE TABLE authors_hash ( name varchar(20), id bigint );
+CREATE TABLE authors_range ( name varchar(20), id bigint );
+CREATE TABLE authors_reference ( name varchar(20), id bigint );
 
 -- this table is used in router executor tests
 CREATE TABLE articles_single_shard_hash (LIKE articles_hash);
@@ -26,13 +41,14 @@ CREATE TABLE articles_single_shard_hash (LIKE articles_hash);
 SELECT master_create_distributed_table('articles_hash', 'author_id', 'hash');
 SELECT master_create_distributed_table('articles_single_shard_hash', 'author_id', 'hash');
 
-
 -- test when a table is distributed but no shards created yet
 SELECT count(*) from articles_hash;
 
 SELECT master_create_worker_shards('articles_hash', 2, 1);
 SELECT master_create_worker_shards('articles_single_shard_hash', 1, 1);
 
+SELECT create_reference_table('authors_reference');
+
 -- create a bunch of test data
 INSERT INTO articles_hash VALUES ( 1,  1, 'arsenous', 9572);
 INSERT INTO articles_hash VALUES ( 2,  2, 'abducing', 13642);
@@ -207,8 +223,7 @@ WITH RECURSIVE hierarchy as (
 SELECT * FROM hierarchy WHERE LEVEL <= 2;
 
 -- logically wrong query, query involves different shards
--- from the same table, but still router plannable due to
--- shard being placed on the same worker.
+-- from the same table
 WITH RECURSIVE hierarchy as (
 	SELECT *, 1 AS level
 		FROM company_employees
@@ -221,6 +236,29 @@ WITH RECURSIVE hierarchy as (
 				ce.company_id = 2))
 SELECT * FROM hierarchy WHERE LEVEL <= 2;
 
+-- CTE with queries other than SELECT is not supported
+WITH new_article AS (
+    INSERT INTO articles_hash VALUES (1,  1, 'arsenous', 9572) RETURNING *
+)
+SELECT * FROM new_article;
+
+-- Modifying statement in nested CTE case is covered by PostgreSQL itself
+WITH new_article AS (
+    WITH nested_cte AS (
+        INSERT INTO articles_hash VALUES (1,  1, 'arsenous', 9572) RETURNING *
+    )
+    SELECT * FROM nested_cte
+)
+SELECT * FROM new_article;
+
+-- Modifying statement in a CTE in subquery is also covered by PostgreSQL
+SELECT * FROM (
+    WITH new_article AS (
+        INSERT INTO articles_hash VALUES (1,  1, 'arsenous', 9572) RETURNING *
+    )
+    SELECT * FROM new_article
+) AS subquery_cte;
+
 -- grouping sets are supported on single shard
 SELECT
 	id, substring(title, 2, 1) AS subtitle, count(*)
@@ -243,13 +281,6 @@ SELECT * FROM articles_hash, position('om' in 'Thomas') WHERE author_id = 1 or a
 -- they are not supported if multiple workers are involved
 SELECT * FROM articles_hash, position('om' in 'Thomas') WHERE author_id = 1 or author_id = 2;
 
--- subqueries are not supported in WHERE clause in Citus
-SELECT * FROM articles_hash WHERE author_id IN (SELECT id FROM authors_hash WHERE name LIKE '%a');
-
-SELECT * FROM articles_hash WHERE author_id IN (SELECT author_id FROM articles_hash WHERE author_id = 1 or author_id = 3);
-
-SELECT * FROM articles_hash WHERE author_id = (SELECT 1);
-
 -- unless the query can be transformed into a join
 SELECT * FROM articles_hash
 WHERE author_id IN (SELECT author_id FROM articles_hash WHERE author_id = 2)
@@ -575,6 +606,18 @@ SELECT a.author_id as first_author, b.word_count as second_word_count
 	FROM articles_hash a, articles_single_shard_hash b
 	WHERE a.author_id = 10 and a.author_id = b.author_id and int4eq(1, 2);
 
+-- partition_column is null clause does not prune out any shards,
+-- all shards remain after shard pruning, not router plannable
+SELECT * 
+	FROM articles_hash a
+	WHERE a.author_id is null;
+
+-- partition_column equals to null clause prunes out all shards
+-- no shards after shard pruning, router plannable
+SELECT * 
+	FROM articles_hash a
+	WHERE a.author_id = null;
+
 -- stable function returning bool
 SELECT * 
 	FROM articles_hash a
@@ -650,6 +693,174 @@ SELECT word_count, rank() OVER (PARTITION BY author_id ORDER BY word_count)
 	FROM articles_hash 
 	WHERE author_id = 1 and 1=0;
 
+-- function calls in WHERE clause with non-relational arguments
+SELECT author_id FROM articles_hash
+	WHERE 
+		substring('hello world', 1, 5) = 'hello'
+	ORDER BY
+		author_id
+	LIMIT 1;
+
+-- when expression evaluates to false
+SELECT author_id FROM articles_hash
+	WHERE 
+		substring('hello world', 1, 4) = 'hello'
+	ORDER BY
+		author_id
+	LIMIT 1;
+
+	
+-- verify range partitioned tables can be used in router plannable queries
+-- just 4 shards to be created for each table to make sure 
+-- they are 'co-located' pairwise
+SET citus.shard_replication_factor TO 1;
+SELECT master_create_distributed_table('authors_range', 'id', 'range');
+SELECT master_create_distributed_table('articles_range', 'author_id', 'range');
+
+SELECT master_create_empty_shard('authors_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 1, shardmaxvalue=10 WHERE shardid = :shard_id;
+
+SELECT master_create_empty_shard('authors_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 11, shardmaxvalue=30 WHERE shardid = :shard_id;
+
+SELECT master_create_empty_shard('authors_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 21, shardmaxvalue=40 WHERE shardid = :shard_id;
+
+SELECT master_create_empty_shard('authors_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 31, shardmaxvalue=40 WHERE shardid = :shard_id;
+
+SELECT master_create_empty_shard('articles_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 1, shardmaxvalue=10 WHERE shardid = :shard_id;
+
+SELECT master_create_empty_shard('articles_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 11, shardmaxvalue=30 WHERE shardid = :shard_id;
+
+SELECT master_create_empty_shard('articles_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 21, shardmaxvalue=40 WHERE shardid = :shard_id;
+
+SELECT master_create_empty_shard('articles_range') as shard_id \gset
+UPDATE pg_dist_shard SET shardminvalue = 31, shardmaxvalue=40 WHERE shardid = :shard_id;
+
+-- single shard select queries are router plannable
+SELECT * FROM articles_range where author_id = 1;
+SELECT * FROM articles_range where author_id = 1 or author_id = 5;
+
+-- zero shard select query is router plannable
+SELECT * FROM articles_range where author_id = 1 and author_id = 2;
+
+-- single shard joins on range partitioned table are router plannable
+SELECT * FROM articles_range ar join authors_range au on (ar.author_id = au.id) 
+	WHERE ar.author_id = 1;
+
+-- zero shard join is router plannable
+SELECT * FROM articles_range ar join authors_range au on (ar.author_id = au.id)
+	WHERE ar.author_id = 1 and au.id = 2;
+
+-- multi-shard join is not router plannable
+SELECT * FROM articles_range ar join authors_range au on (ar.author_id = au.id)
+	WHERE ar.author_id = 35;
+
+-- this is a bug, it is a single shard join query but not router plannable
+SELECT * FROM articles_range ar join authors_range au on (ar.author_id = au.id) 
+	WHERE ar.author_id = 1 or au.id = 5;
+
+-- bogus query, join on non-partition column, but router plannable due to filters
+SELECT * FROM articles_range ar join authors_range au on (ar.id = au.id) 
+	WHERE ar.author_id = 1 and au.id < 10;
+
+-- join between hash and range partition tables are router plannable
+-- only if both tables pruned down to single shard and co-located on the same
+-- node.
+-- router plannable
+SELECT * FROM articles_hash ar join authors_range au on (ar.author_id = au.id)
+	WHERE ar.author_id = 2;
+
+-- not router plannable
+SELECT * FROM articles_hash ar join authors_range au on (ar.author_id = au.id)
+	WHERE ar.author_id = 3;
+
+-- join between a range partitioned table and reference table is router plannable
+SELECT * FROM articles_range ar join authors_reference au on (ar.author_id = au.id)
+	WHERE ar.author_id = 1;
+
+-- still hits a single shard and router plannable
+SELECT * FROM articles_range ar join authors_reference au on (ar.author_id = au.id)
+	WHERE ar.author_id = 1 or ar.author_id = 5;
+
+-- it is not router plannable if hit multiple shards
+SELECT * FROM articles_range ar join authors_reference au on (ar.author_id = au.id)
+	WHERE ar.author_id = 1 or ar.author_id = 15;
+	
+-- following is a bug, function should have been
+-- evaluated at master before going to worker
+-- need to use a append distributed table here
+SELECT master_create_distributed_table('articles_append', 'author_id', 'append');
+SET citus.shard_replication_factor TO 1;
+SELECT master_create_empty_shard('articles_append') AS shard_id \gset
+UPDATE pg_dist_shard SET shardmaxvalue = 100, shardminvalue=1 WHERE shardid = :shard_id;
+
+SELECT author_id FROM articles_append
+	WHERE 
+		substring('articles_append'::regclass::text, 1, 5) = 'hello'
+	ORDER BY
+		author_id
+	LIMIT 1;
+
+-- same query with where false but evaluation left to worker
+SELECT author_id FROM articles_append
+	WHERE 
+		substring('articles_append'::regclass::text, 1, 4) = 'hello'
+	ORDER BY
+		author_id
+	LIMIT 1;
+
+-- same query on router planner with where false but evaluation left to worker
+SELECT author_id FROM articles_single_shard_hash
+	WHERE 
+		substring('articles_single_shard_hash'::regclass::text, 1, 4) = 'hello'
+	ORDER BY
+		author_id
+	LIMIT 1;
+
+SELECT author_id FROM articles_hash
+	WHERE 
+		author_id = 1
+		AND substring('articles_hash'::regclass::text, 1, 5) = 'hello'
+	ORDER BY
+		author_id
+	LIMIT 1;
+
+-- create a dummy function to be used in filtering
+CREATE OR REPLACE FUNCTION someDummyFunction(regclass)
+    RETURNS text AS
+$$
+BEGIN
+    RETURN md5($1::text);
+END;
+$$ LANGUAGE 'plpgsql' IMMUTABLE;
+
+-- not router plannable, returns all rows 
+SELECT * FROM articles_hash
+	WHERE
+		someDummyFunction('articles_hash') = md5('articles_hash')
+	ORDER BY
+		author_id, id
+	LIMIT 5;
+
+-- router plannable, errors
+SELECT * FROM articles_hash
+	WHERE
+		someDummyFunction('articles_hash') = md5('articles_hash') AND author_id = 1
+	ORDER BY
+		author_id, id
+	LIMIT 5;
+
+-- temporarily turn off debug messages before dropping the function
+SET client_min_messages TO 'NOTICE';
+DROP FUNCTION someDummyFunction(regclass);
+
+SET client_min_messages TO 'DEBUG2';
+
 -- complex query hitting a single shard 	
 SELECT
 	count(DISTINCT CASE
@@ -688,6 +899,15 @@ SELECT *
 	ORDER BY id;
 END;
 
+-- queries inside read-only transactions can be router plannable
+BEGIN;
+SET TRANSACTION READ ONLY;
+SELECT *
+	FROM articles_hash
+	WHERE author_id = 1
+	ORDER BY id;
+END;
+
 -- cursor queries are router plannable
 BEGIN;
 DECLARE test_cursor CURSOR FOR 
@@ -698,6 +918,7 @@ DECLARE test_cursor CURSOR FOR
 FETCH test_cursor;
 FETCH ALL test_cursor;
 FETCH test_cursor; -- fetch one row after the last
+FETCH BACKWARD test_cursor;
 END;
 
 -- queries inside copy can be router plannable
@@ -719,7 +940,7 @@ SELECT count(*), count(*) FILTER (WHERE id < 3)
 	FROM articles_hash
 	WHERE author_id = 1;
 
--- non-router plannable queries do not support filters
+-- non-router plannable queries also support filters
 SELECT count(*), count(*) FILTER (WHERE id < 3)
 	FROM articles_hash
 	WHERE author_id = 1 or author_id = 2;
@@ -769,16 +990,15 @@ $$ LANGUAGE plpgsql;
 SELECT * FROM author_articles_id_word_count();
 
 -- materialized views can be created for router plannable queries
-CREATE MATERIALIZED VIEW mv_articles_hash AS
+CREATE MATERIALIZED VIEW mv_articles_hash_empty AS
 	SELECT * FROM articles_hash WHERE author_id = 1;
+SELECT * FROM mv_articles_hash_empty;
 
-SELECT * FROM mv_articles_hash;
-
-CREATE MATERIALIZED VIEW mv_articles_hash_error AS
+CREATE MATERIALIZED VIEW mv_articles_hash_data AS
 	SELECT * FROM articles_hash WHERE author_id in (1,2);
-	
--- router planner/executor is disabled for task-tracker executor
--- following query is router plannable, but router planner is disabled
+SELECT * FROM mv_articles_hash_data;
+
+-- router planner/executor is now enabled for task-tracker executor
 SET citus.task_executor_type to 'task-tracker';
 SELECT id
 	FROM articles_hash
@@ -794,12 +1014,55 @@ SELECT id
 
 SET client_min_messages to 'NOTICE';
 
+-- test that a connection failure marks placements invalid
+SET citus.shard_replication_factor TO 2;
+CREATE TABLE failure_test (a int, b int);
+SELECT master_create_distributed_table('failure_test', 'a', 'hash');
+SELECT master_create_worker_shards('failure_test', 2);
+
+CREATE USER router_user;
+GRANT INSERT ON ALL TABLES IN SCHEMA public TO router_user;
+\c - - - :worker_1_port
+CREATE USER router_user;
+GRANT INSERT ON ALL TABLES IN SCHEMA public TO router_user;
+\c - router_user - :master_port
+-- first test that it is marked invalid inside a transaction block
+-- we will fail to connect to worker 2, since the user does not exist
+BEGIN;
+INSERT INTO failure_test VALUES (1, 1);
+SELECT shardid, shardstate, nodename, nodeport FROM pg_dist_shard_placement
+	WHERE shardid IN (
+		SELECT shardid FROM pg_dist_shard
+		WHERE logicalrelid = 'failure_test'::regclass
+	)
+	ORDER BY placementid;
+ROLLBACK;
+INSERT INTO failure_test VALUES (2, 1);
+SELECT shardid, shardstate, nodename, nodeport FROM pg_dist_shard_placement
+	WHERE shardid IN (
+		SELECT shardid FROM pg_dist_shard
+		WHERE logicalrelid = 'failure_test'::regclass
+	)
+	ORDER BY placementid;
+\c - postgres - :worker_1_port
+DROP OWNED BY router_user;
+DROP USER router_user;
+\c - - - :master_port
+DROP OWNED BY router_user;
+DROP USER router_user;
+DROP TABLE failure_test;
+
 DROP FUNCTION author_articles_max_id();
 DROP FUNCTION author_articles_id_word_count();
 
-DROP MATERIALIZED VIEW mv_articles_hash;
+DROP MATERIALIZED VIEW mv_articles_hash_empty;
+DROP MATERIALIZED VIEW mv_articles_hash_data;
 
 DROP TABLE articles_hash;
 DROP TABLE articles_single_shard_hash;
 DROP TABLE authors_hash;
+DROP TABLE authors_range;
+DROP TABLE authors_reference;
 DROP TABLE company_employees;
+DROP TABLE articles_range;
+DROP TABLE articles_append;
