@@ -5,7 +5,6 @@
 
 
 ALTER SEQUENCE pg_catalog.pg_dist_shardid_seq RESTART 870000;
-ALTER SEQUENCE pg_catalog.pg_dist_jobid_seq RESTART 870000;
 
 
 CREATE TABLE testtableddl(somecol int, distributecol text NOT NULL);
@@ -20,22 +19,33 @@ ALTER TABLE testtableddl ALTER COLUMN distributecol TYPE text;
 -- verify that the distribution column can't be dropped
 ALTER TABLE testtableddl DROP COLUMN distributecol;
 
--- verify that the table cannot be dropped in a transaction block
+-- verify that the table can be dropped in a transaction block
 \set VERBOSITY terse
 BEGIN;
 DROP TABLE testtableddl;
-ROLLBACK;
+COMMIT;
 \set VERBOSITY default
 
+-- recreate testtableddl
+CREATE TABLE testtableddl(somecol int, distributecol text NOT NULL);
+SELECT master_create_distributed_table('testtableddl', 'distributecol', 'append');
+
 -- verify that the table can be dropped
 DROP TABLE testtableddl;
 
 -- verify that the table can dropped even if shards exist
 CREATE TABLE testtableddl(somecol int, distributecol text NOT NULL);
+
+-- create table and do create empty shard test here, too
+SET citus.shard_replication_factor TO 1;
 SELECT master_create_distributed_table('testtableddl', 'distributecol', 'append');
 SELECT 1 FROM master_create_empty_shard('testtableddl');
+
+-- now actually drop table and shards
 DROP TABLE testtableddl;
 
+RESET citus.shard_replication_factor;
+
 -- ensure no metadata of distributed tables are remaining
 SELECT * FROM pg_dist_partition;
 SELECT * FROM pg_dist_shard;
